import json
import os
import re
from datetime import datetime
from enum import Enum

from flask import url_for
from sqlalchemy import sql
from sqlalchemy.orm.exc import NoResultFound

# TODO: circular import here...
import Pegasus.db as db
from Pegasus import user


def timestamp(dt):
    return (dt - datetime(1970, 1, 1)).total_seconds()


class EMError(Exception):
    def __init__(self, message, status_code=500):
        Exception.__init__(self, message)
        self.status_code = status_code


def validate_ensemble_name(name):
    if name is None:
        raise EMError("Specify ensemble name")
    if len(name) >= 100:
        raise EMError("Ensemble name too long: %d" % len(name))
    if re.match(r"\A[a-zA-Z0-9_-]+\Z", name) is None:
        raise EMError("Invalid ensemble name: %s" % name)
    return name


def validate_priority(priority):
    try:
        return int(priority)
    except ValueError:
        raise EMError("Invalid priority: %s" % priority)


class EnsembleBase:
    def set_name(self, name):
        self.name = validate_ensemble_name(name)

    def set_created(self):
        self.created = datetime.utcnow()

    def set_updated(self):
        self.updated = datetime.utcnow()


class States(set):
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError


EnsembleStates = States(["ACTIVE", "HELD", "PAUSED"])
EnsembleWorkflowStates = States(
    [
        "READY",
        "PLANNING",
        "PLAN_FAILED",
        "QUEUED",
        "RUN_FAILED",
        "RUNNING",
        "FAILED",
        "SUCCESSFUL",
        "ABORTED",
    ]
)


class Ensemble(EnsembleBase):
    def __init__(self, username, name):
        self.username = username
        self.set_name(name)
        self.set_created()
        self.set_updated()
        self.state = EnsembleStates.ACTIVE
        self.set_max_running(1)
        self.set_max_planning(1)

    def set_state(self, state):
        state = state.upper()
        if state not in EnsembleStates:
            raise EMError("Invalid ensemble state: %s" % state)
        self.state = state

    def set_max_running(self, max_running):
        try:
            max_running = int(max_running)
            if max_running < 1:
                raise EMError("Value for max_running must be >= 1: %s" % max_running)
            self.max_running = max_running
        except ValueError:
            raise EMError("Invalid value for max_running: %s" % max_running)

    def set_max_planning(self, max_planning):
        try:
            max_planning = int(max_planning)
            if max_planning < 1:
                raise EMError("Value for max_planning must be >= 1: %s" % max_planning)
            self.max_planning = max_planning
        except ValueError:
            raise EMError("Invalid value for max_planning: %s" % max_planning)

    def get_localdir(self):
        u = user.get_user_by_username(self.username)
        edir = u.get_ensembles_dir()
        return os.path.join(edir, self.name)

    def get_object(self):
        return {
            "id": self.id,
            "name": self.name,
            "created": timestamp(self.created),
            "updated": timestamp(self.updated),
            "state": self.state,
            "max_running": self.max_running,
            "max_planning": self.max_planning,
            "workflows": url_for(
                "route_list_ensemble_workflows", name=self.name, _external=True
            ),
            "href": url_for("route_get_ensemble", name=self.name, _external=True),
        }


class EnsembleWorkflow(EnsembleBase):
    def __init__(self, ensemble_id, name, basedir, plan_command):
        self.ensemble_id = ensemble_id
        self.set_name(name)
        self.set_basedir(basedir)
        self.set_created()
        self.set_updated()
        self.state = EnsembleWorkflowStates.READY
        self.set_priority(0)
        self.set_plan_command(plan_command)
        self.set_wf_uuid(None)
        self.set_submitdir(None)

    def set_state(self, state):
        state = state.upper()
        if state not in EnsembleWorkflowStates:
            raise EMError("Invalid ensemble workflow state: %s" % state)
        self.state = state

    def change_state(self, state):
        # The allowed transitions are:
        #   PLAN_FAILED -> READY
        #   RUN_FAILED -> QUEUED
        #   RUN_FAILED -> READY
        #   FAILED -> QUEUED
        #   FAILED -> READY
        if self.state == EnsembleWorkflowStates.PLAN_FAILED:
            if state != EnsembleWorkflowStates.READY:
                raise EMError("Can only replan workflows in PLAN_FAILED state")
        elif self.state == EnsembleWorkflowStates.RUN_FAILED:
            if state not in (
                EnsembleWorkflowStates.READY,
                EnsembleWorkflowStates.QUEUED,
            ):
                raise EMError("Can only replan or rerun workflows in RUN_FAILED state")
        elif self.state == EnsembleWorkflowStates.FAILED:
            if state not in (
                EnsembleWorkflowStates.READY,
                EnsembleWorkflowStates.QUEUED,
            ):
                raise EMError("Can only replan or rerun workflows in FAILED state")
        else:
            raise EMError("Invalid state change: {} -> {}".format(self.state, state))

        self.set_state(state)

    def set_priority(self, priority):
        self.priority = validate_priority(priority)

    def set_wf_uuid(self, wf_uuid):
        if wf_uuid is not None and len(wf_uuid) != 36:
            raise EMError("Invalid wf_uuid")
        self.wf_uuid = wf_uuid

    def set_basedir(self, basedir):
        self.basedir = basedir

    def set_submitdir(self, submitdir):
        self.submitdir = submitdir

    def set_plan_command(self, plan_command):
        self.plan_command = plan_command

    def _get_file(self, suffix):
        edir = self.ensemble.get_localdir()
        wf = self.name
        filename = "{}.{}".format(wf, suffix)
        return os.path.join(edir, filename)

    def get_basedir(self):
        return self.basedir

    def get_pidfile(self):
        return self._get_file("plan.pid")

    def get_resultfile(self):
        return self._get_file("plan.result")

    def get_runfile(self):
        return self._get_file("plan.run")

    def get_logfile(self):
        return self._get_file("log")

    def get_plan_command(self):
        return self.plan_command

    def get_object(self):
        return {
            "id": self.id,
            "name": self.name,
            "created": timestamp(self.created),
            "updated": timestamp(self.updated),
            "state": self.state,
            "priority": self.priority,
            "wf_uuid": self.wf_uuid,
            "href": url_for(
                "route_get_ensemble_workflow",
                ensemble=self.ensemble.name,
                workflow=self.name,
                _external=True,
            ),
        }

    def get_detail_object(self):
        o = self.get_object()
        o["basedir"] = self.basedir
        o["plan_command"] = self.plan_command
        o["log"] = self.get_logfile()
        o["submitdir"] = self.submitdir
        return o


class Ensembles:
    def __init__(self, session):
        self.session = session

    def list_ensembles(self, username):
        q = self.session.query(Ensemble)
        q = q.filter(Ensemble.username == username)
        q = q.order_by(Ensemble.created)
        return q.all()

    def list_actionable_ensembles(self):
        states = (
            EnsembleWorkflowStates.READY,
            EnsembleWorkflowStates.PLANNING,
            EnsembleWorkflowStates.QUEUED,
            EnsembleWorkflowStates.RUNNING,
        )
        stmt = (
            sql.exists()
            .where(Ensemble.id == EnsembleWorkflow.ensemble_id)
            .where(EnsembleWorkflow.state.in_(states))
        )
        return self.session.query(Ensemble).filter(stmt).all()

    def get_ensemble(self, username, name):
        try:
            return (
                self.session.query(Ensemble)
                .filter(Ensemble.username == username, Ensemble.name == name)
                .one()
            )
        except NoResultFound:
            raise EMError("No such ensemble: %s" % name, 404)

    def get_ensemble_name(self, ensemble_id: int):
        """Given an ensemble id, get its name

        :param ensemble_id: id of target ensemble
        :type ensemble_id: int
        :return: name of target ensemble
        :rtype: str
        """
        return self.session.query(Ensemble).filter_by(id=ensemble_id).first().name

    def create_ensemble(self, username, name, max_running, max_planning):
        if (
            self.session.query(Ensemble)
            .filter(Ensemble.username == username, Ensemble.name == name)
            .count()
            > 0
        ):
            raise EMError("Ensemble %s already exists" % name, 400)

        ensemble = Ensemble(username, name)
        ensemble.set_max_running(max_running)
        ensemble.set_max_planning(max_planning)
        self.session.add(ensemble)
        self.session.flush()
        return ensemble

    def list_ensemble_workflows(self, ensemble_id):
        q = self.session.query(EnsembleWorkflow)
        q = q.filter(EnsembleWorkflow.ensemble_id == ensemble_id)
        q = q.order_by(EnsembleWorkflow.created)
        return q.all()

    def get_ensemble_workflow(self, ensemble_id, name):
        try:
            q = self.session.query(EnsembleWorkflow)
            q = q.filter(
                EnsembleWorkflow.ensemble_id == ensemble_id,
                EnsembleWorkflow.name == name,
            )
            return q.one()
        except NoResultFound:
            raise EMError("No such ensemble workflow: %s" % name, 404)

    def create_ensemble_workflow(
        self, ensemble_id, name, basedir, priority, plan_command
    ):

        # Verify that the workflow doesn't already exist
        q = self.session.query(EnsembleWorkflow)
        q = q.filter(
            EnsembleWorkflow.ensemble_id == ensemble_id, EnsembleWorkflow.name == name
        )
        if q.count() > 0:
            raise EMError("Ensemble workflow %s already exists" % name, 400)

        # Create database record
        w = EnsembleWorkflow(ensemble_id, name, basedir, plan_command)
        w.set_priority(priority)
        self.session.add(w)
        self.session.flush()

        return w

    def write_planning_script(
        self,
        f,
        basedir,
        bundledir,
        dax,
        sites,
        output_site,
        staging_sites=None,
        clustering=None,
        force=False,
        cleanup=None,
    ):

        f.write("#!/bin/bash\n")
        f.write("pegasus-plan \\\n")

        # We need to make sure that the dashboard info is
        # sent to the same database we are using
        f.write("-Dpegasus.dashboard.output=%s \\\n" % self.dburi)

        f.write("--conf pegasus.properties \\\n")
        f.write("--site %s \\\n" % ",".join(sites))
        f.write("--output-site %s \\\n" % output_site)

        if staging_sites is not None and len(staging_sites) > 0:
            pairs = ["{}={}".format(k, v) for k, v in staging_sites.items()]
            f.write("--staging-site %s \\\n" % ",".join(pairs))

        if clustering is not None and len(clustering) > 0:
            f.write("--cluster %s \\\n" % ",".join(clustering))

        if force:
            f.write("--force \\\n")

        if cleanup is not None:
            f.write("--cleanup %s \\\n" % cleanup)

        f.write("--dir %s \\\n" % os.path.join(basedir, "submit"))
        f.write("--dax %s \\\n" % os.path.join(bundledir, dax))
        f.write("--input-dir %s \n" % bundledir)

        f.write("exit $?")


# --- workflow trigger ---------------------------------------------------------
TriggerStates = States(["READY", "RUNNING", "FAILED", "STOPPED"])


class TriggerType(Enum):
    """Supported trigger types"""

    CRON = "CRON"
    FILE_PATTERN = "FILE_PATTERN"


class Triggers:
    """Trigger table data access object"""

    def __init__(self, session):
        self.session = session

    def get_trigger(self, ensemble_id: int, trigger_name: str):
        """Get a specific trigger

        :param ensemble_id: id of the nsemble that the trigger belongs to
        :type ensemble_id: int
        :param trigger_name: name of the trigger
        :type trigger_name: str
        """
        try:
            return (
                self.session.query(db.schema.Trigger)
                .filter_by(ensemble_id=ensemble_id, name=trigger_name)
                .one()
            )
        except NoResultFound:
            raise EMError(
                "No such trigger: {} assigned to ensemble id: {}".format(
                    trigger_name, ensemble_id
                ),
                status_code=404,
            )

    def list_triggers(self):
        """List all triggers"""
        return self.session.query(db.schema.Trigger).all()

    def list_triggers_by_ensemble(self, username: str, ensemble: str):
        """List all triggers belonging to a specific ensemble

        :param username: name of user making request
        :type username: str
        :param ensemble: ensemble name
        :type ensemble: str
        :return: list of Triggers
        :rtype: List[Trigger]
        """
        q = self.session.query(db.schema.Trigger).filter(
            Ensemble.username == username,
            Ensemble.name == ensemble,
            Ensemble.id == db.schema.Trigger.ensemble_id,
        )

        return q.all()

    def insert_trigger(
        self,
        ensemble_id: int,
        trigger: str,
        trigger_type: str,
        workflow_script: str,
        workflow_args: str,
        **trigger_kwargs
    ):
        """Insert a trigger

        :param ensemble_id: the id of the ensemble this trigger belongs to
        :type ensemble_id: int
        :param trigger: name of the trigger
        :type trigger: str
        :param trigger_type: the type of the trigger (e.g. CRON, FILE_PATTERN, etc.)
        :type trigger_type: str
        :param workflow_script: the workflow generator & planning script
        :type workflow_script: str
        :param workflow_args: any command line args to be pass to the workflow_script
        :type workflow_args: str
        :param trigger_kwargs: any arguments specific to the trigger (e.g. interval=10s)
        """

        self.session.add(
            db.schema.Trigger(
                ensemble_id=ensemble_id,
                name=trigger,
                state="READY",
                workflow=json.dumps({"script": workflow_script, "args": workflow_args}),
                args=json.dumps(trigger_kwargs),
                _type=trigger_type,
            )
        )

        self.session.commit()

    def update_state(self, ensemble_id: int, trigger_id: int, new_state: str):
        """Update the state of a trigger

        :param ensemble_id: id of the ensemble to which the trigger belongs
        :type ensemble_id: int
        :param trigger_id: id of the trigger
        :type trigger_id: int
        :param new_state: the new state the trigger will be updated to
        :type new_state: str
        """

        self.session.query(db.schema.Trigger).filter_by(
            ensemble_id=ensemble_id, _id=trigger_id
        ).update({"state": new_state})

        self.session.commit()

    def delete_trigger(self, ensemble_id: int, trigger: str):
        """Delete a trigger from the database

        :param ensemble_id: id of the ensemble to which the trigger belongs
        :type ensemble: int
        :param trigger: name of the trigger
        :type trigger: str
        """

        self.session.query(db.schema.Trigger).filter_by(
            ensemble_id=ensemble_id, name=trigger
        ).delete()

        self.session.commit()

    @staticmethod
    def get_object(trigger):
        """Get trigger as dict"""

        return {
            "id": trigger._id,
            "ensemble_id": trigger.ensemble_id,
            "name": trigger.name,
            "state": trigger.state,
            "workflow": json.loads(trigger.workflow),
            "args": json.loads(trigger.args) if trigger.args else None,
            "type": trigger._type,
        }



import calendar
import re
import time

from . import magicdate

ISO_DATE_PARTS = re.compile(
    r"(\d\d\d\d)(?:-(\d\d)(?:-(\d\d)(?:T(\d\d)(?::(\d\d)(?::(\d\d)(?:\.(\d+))?)?)?)?)?)?(Z|[+-]\d\d:\d\d)?"
)
ISO_DATE_ZEROES = (None, "01", "01", "00", "00", "00", "0")

NUMBER_DATE = re.compile(r"(\d+)(?:\.(\d+))?")

# Date format constants
UNKNOWN = "Unknown"
ISO8601 = "ISO8601 date"
ENGLISH = "Natural English language date"
SECONDS = "Seconds since UNIX epoch"

DATE_FMT = "%04d-%02d-%02dT%02d:%02d:%02d"


class DateFormatError(Exception):
    pass


def getLocaltimeOffsetSec(t=None):
    """Return current localtime offset at time 't' (default=now)
    from UTC as a number of seconds.
    """
    if t is None:
        t = time.time()
    # this doesn't handle DST properly
    # offs_sec = time.mktime(time.localtime(t)) - time.mktime(time.gmtime(t))
    # this does:
    offs_sec = calendar.timegm(time.localtime(t)) - time.mktime(time.localtime(t))
    return offs_sec


def getLocaltimeOffsetParts(t=None):
    """Return current localtime offset at time 't' (default=now)
    from GMT as a triple:
       hours, minutes, sign (1 for positive or 0, -1 for negative)
    """
    offs_sec = getLocaltimeOffsetSec(t)
    offs_hr = int(offs_sec / 3600)
    offs_min = int((offs_sec - (offs_hr * 3600)) / 60)
    sign = (1, -1)[offs_hr < 0]
    return abs(offs_hr), abs(offs_min), sign


def getLocaltimeISO(t):
    """Get an ISO8601 string for the local timezone at time 't', where
    t is either seconds since the epoch as a float, or a tuple like the one
    returned by time.localtime(), or a string date

    Return the string, which may be [+-]HH:MM or 'Z'.
    """
    # If 't' is a list, tuple or string, first convert to seconds
    if isinstance(t, list) or isinstance(t, tuple) or isinstance(t, str):
        # Convert a string to a list
        if isinstance(t, str):
            parts = splitISODate(t)
        else:
            # If it's a tuple or list, just assign
            parts = t
        # From the tuple or list, calculate time offset for the
        # same year, month, day, and hour.
        tm = (
            int(parts[0]),
            int(parts[1]),
            int(parts[2]),
            int(parts[3]),
            0,
            0,
            0,
            1,
            -1,
        )
        t = time.mktime(tm)
    # Given the time, figure out the timezone offset
    hr, min, sign = getLocaltimeOffsetParts(t)
    # Format the offset as an ISO string
    if hr == 0 and min == 0:
        s = "Z"
    else:
        s = "%s%02d:%02d" % (("", "+", "-")[sign], hr, min)
    return s


def splitISODate(s):
    """Split an ISO date into parts like those returned by localtime()

    Raises a DateFormatError if the date couldn't be split.

    Returns the list of parts.
    """
    m = ISO_DATE_PARTS.match(s)
    if not m:
        raise DateFormatError("Not a partial ISO date: %s" % s)
    parts = list(m.groups())
    # Fill in missing parts with 'zero'
    for i, part in enumerate(parts[:-1]):
        if part is None:
            parts[i] = ISO_DATE_ZEROES[i]
    return parts


def completeISO(s, is_gmt=False, set_gmt=False):
    """Make a partial ISO8601 date into a full ISO8601 date.

    If 'gmt' is True, assume timezone is GMT when not given.
    Otherwise, assume localtime.
    """
    parts = splitISODate(s)
    # add timezone
    iso_str = None
    if parts[7] is None:
        if is_gmt and set_gmt:
            tz = "Z"
        elif not is_gmt and not set_gmt:
            tz = getLocaltimeISO(parts)
        else:
            # adjust time GMT to localtime or localtime to GMT
            # easiest at this point to just format string from time
            p = list(map(int, parts[:-1] + [0, -1]))
            if set_gmt:
                t = time.mktime(p)
                iso_str = utcFormatISO(t)
            else:
                t = calendar.timegm(p)
                iso_str = localtimeFormatISO(t)
    else:
        # explicit timezone overrides
        tz = parts[7]
    if iso_str is None:
        iso_str = "-".join(parts[:3]) + "T" + ":".join(parts[3:6]) + "." + parts[6] + tz
    return iso_str


def parseISO(s):
    """Parse ISO8601 (string) date into a floating point seconds since epoch UTC.

    The string must be an ISO8601 date of the form
          YYYY-MM-DDTHH:MM:SS[.fff...](Z|[+-]dd:dd)

    If something doesn't parse, a DateFormatError will be raised.

    The return value is floating point seconds since the UNIX
    epoch (January 1, 1970 at midnight UTC).
    """
    # if it's too short
    if len(s) < 7:
        raise DateFormatError("Date '%s' is too short" % s)
    # UTC timezone?
    if s[-1] == "Z":
        tz_offs, tz_len = 0, 1
    # explicit +/-nn:nn timezone?
    elif s[-3] == ":":
        tz_offs = int(s[-6] + "1") * (int(s[-5:-3]) * 3600 + int(s[-2:]) * 60)
        tz_len = 6
    # otherwise
    else:
        raise DateFormatError("Date '%s' is missing timezone" % s)
    # split into components
    cal, clock = s.split("T")
    year, month, day = cal.split("-")
    hr, minute, sec = clock[:-tz_len].split(":")
    # handle fractional seconds
    frac = 0
    point = sec.find(".")
    if point != -1:
        frac = float(sec[point + 1 :]) / pow(10, len(sec) - point - 1)
        sec = sec[:point]
    # use calendar to get seconds since epoch
    args = list(map(int, (year, month, day, hr, minute, sec))) + [0, 1, -1]
    return calendar.timegm(args) + frac - tz_offs  # adjust to GMT


def makeISO(value, is_gmt=False, set_gmt=False):
    """If value is a tuple, assume it is the one returned by time.gmtime() or time.localtime()
    Otherwise, assume value is an English language description (for partial ISO
    strings, use completeISO() instead).

    Return an ISO8601 string, with timezone set to GMT or localtime.
    """
    tz_str = "Z"  # assume GMT
    if isinstance(value, tuple) or isinstance(value, list):
        fmt = ("%04d", "-%02d", "-%02d", "T%02d", ":%02d", ":%02d")
        s = "".join(f % v for f, v in zip(fmt, value))
        if not gmt:
            tz_str = getLocaltimeISO(value)
        iso = s + tz_str
    else:
        try:
            d = magicdate.magicdate(value)
        except Exception:
            raise ValueError("magicdate cannot parse '%s'" % value)
        partial_iso = d.isoformat()
        iso = completeISO(partial_iso, is_gmt=is_gmt, set_gmt=set_gmt)
    return iso


def guess(
    s, parse=True, is_gmt=False, set_gmt=False, try_iso=True, try_num=True, try_en=True
):
    """Guess the format, and optionally parse, the input string.

    If 'is_gmt' is True, assume timezone is GMT when not given.
    Otherwise, assume localtime.
    If 'set_gmt' is True then set the timezone to GMT, otherwise
    set it to localtime.

    The answer is a pair containing the guessed format and, if the 'parse'
    flag was given, the parsed value as seconds since the epoch, otherwise None.

    The format is a constant defined in this module:
      UNKNOWN - Cannot guess the format (associated value is None)
      ISO8601 - This is a prefix of the ISO8601 format accepted by completeISO()
      ENGLISH - This is an natural English-language format accepted by makeISO()
      SECONDS - This is seconds since the UNIX epoch (Midnight on 1970/1/1).
    """
    if not (s):
        return UNKNOWN, None
    sec = None
    s = s.strip()
    # try ISO8601
    if try_iso:
        m = ISO_DATE_PARTS.match(s)
        if m and m.start() == 0 and m.end() == len(s):
            if parse:
                if s[-1] == "Z":
                    # explicit timezone overrides option
                    is_gmt = True
                iso_s = completeISO(s, is_gmt=is_gmt, set_gmt=set_gmt)
                sec = parseISO(iso_s)
            return ISO8601, sec
    # try number
    if try_num:
        m = NUMBER_DATE.match(s)
        if m and m.start() == 0 and m.end() == len(s):
            if parse:
                sec = float(s)
            return SECONDS, sec
    # try natural language
    if try_en:
        try:
            d = magicdate.magicdate(s)
        except Exception:
            d = None
        if d is not None:
            if parse:
                partial_iso = d.isoformat()
                iso = completeISO(partial_iso, is_gmt=False, set_gmt=set_gmt)
                sec = parseISO(iso)
            return ENGLISH, sec
    # default: unknown
    return UNKNOWN, None


def utcFormatISO(sec):
    """Format 'sec' seconds since the epoch as a UTC ISO8601 date,
    in the UTC (GMT) timezone.
    Use the format: "YYYY-MM-DDThh:mm:ss.fffffffZ"
    """
    tm = time.gmtime(sec)
    usec = int((sec - int(sec)) * 1e6)
    iso_date = "%s.%06dZ" % (DATE_FMT % tm[0:6], usec)
    return iso_date


def localtimeFormatISO(sec):
    """Format 'sec' seconds since the epoch as a ISO8601 date,
    in the local timezone.
    Use the format: "YYYY-MM-DDThh:mm:ss.fffffff[+-]HH:MM"
    """
    tm = time.localtime(sec)
    usec = int((sec - int(sec)) * 1000000)
    hr, min, sign = getLocaltimeOffsetParts(sec)
    iso_date = "%s.%06d%s%02d:%02d" % (
        DATE_FMT % tm[0:6],
        usec,
        ("-", "+")[sign == 1],
        hr,
        min,
    )
    return iso_date


# Syslog-style dates (always in localtime)
SYSLOG_DATE_RE = re.compile(
    r"\s*(...)\s+(...)\s+(\d\d?) " + r"(\d\d):(\d\d):(\d\d)\s+(\d\d\d\d)\s*"
)
MONTHS = {
    "Jan": 1,
    "Feb": 2,
    "Mar": 3,
    "Apr": 4,
    "May": 5,
    "Jun": 6,
    "Jul": 7,
    "Aug": 8,
    "Sep": 9,
    "Oct": 10,
    "Nov": 11,
    "Dec": 12,
}


def parseSyslogDate(date):
    """Parse syslog-style date to seconds UTC.
    Expected format: Fri Oct 24 04:18:36 2008
    """
    m = SYSLOG_DATE_RE.match(date)
    if m is None:
        raise ValueError("bad syslog date '%s'" % date)
    g = m.groups()
    month = MONTHS[g[1]]
    day, hh, mm, ss, year = list(map(int, g[2:]))
    sec = time.mktime((year, month, day, hh, mm, ss, 0, 0, -1))
    return sec



"""
File wrapper that changes the semantics of the readline function
to never return a non-empty line without a line terminator.
"""
__rcsid__ = "$Id: nlreadline.py 22911 2008-06-03 05:06:10Z dang $"
__author__ = "Dan Gunter"


class BufferedReadline:
    """Change semantics of file.readline() to return either
    a complete line with a newline terminator or an empty line.
    Partial lines are buffered between calls until the newline
    is found.
    """

    def __init__(self, fileobj):
        self._f = fileobj
        self._buf = ""

    def __getattr__(self, x):
        """Delegate all public methods except readline()."""
        if x and x[0] == "_":
            # look up private methods/vars locally and raise
            # a normal-looking AttributeError if not found
            try:
                return self.__dict__[x]
            except KeyError:
                raise AttributeError(
                    "'{}' object has no attribute '{}'".format(
                        self.__class__.__name__, x
                    )
                )
        return getattr(self._f, x)

    def readline(self):
        """Override readline() to get new semantics."""
        if self._f is None:
            return ""
        line = self._f.readline()
        if line == "":
            # empty lines are returned as before
            pass
        elif line[-1] == "\n":
            # complete lines are returned along with
            # the buffered data (if any)
            if self._buf:
                line = self._buf + line
                self._buf = ""
        else:
            # incomplete lines are buffered, and an empty
            # line is returned
            self._buf += line
            line = ""
        return line

    def readlines(self):
        """Override readlines() so it calls our readline()."""
        if self._f is None:
            return ""
        return [line for line in self.readline()]

    def xreadlines(self):
        """Override xreadlines() so it calls our readline()."""
        while True:
            line = self.readline()
            if line:
                yield line
            else:
                raise StopIteration

    def close(self):
        if self._f is not None:
            self._f.close()
        self._f = None



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__author__ = "Rafael Ferreira da Silva"

import getpass
import logging
import os
import subprocess
from sqlite3 import Connection as SQLite3Connection
from stat import ST_MODE
from urllib.parse import urlparse

from sqlalchemy import create_engine, event, exc, orm
from sqlalchemy.engine import Engine

from Pegasus import user as users
from Pegasus.tools import properties, utils

__all__ = ["connect"]

log = logging.getLogger(__name__)

# -------------------------------------------------------------------
# Connection Properties
PROP_CATALOG_ALL_TIMEOUT = "pegasus.catalog.*.timeout"
PROP_CATALOG_ALL_DB_TIMEOUT = "pegasus.catalog.*.db.timeout"
PROP_CATALOG_MASTER_URL = "pegasus.catalog.master.url"
PROP_CATALOG_MASTER_TIMEOUT = "pegasus.catalog.master.timeout"
PROP_CATALOG_MASTER_DB_TIMEOUT = "pegasus.catalog.master.db.timeout"
PROP_CATALOG_REPLICA_DB_URL = "pegasus.catalog.replica.db.url"
PROP_CATALOG_REPLICA_TIMEOUT = "pegasus.catalog.replica.timeout"
PROP_CATALOG_REPLICA_DB_TIMEOUT = "pegasus.catalog.replica.db.timeout"
PROP_CATALOG_WORKFLOW_URL = "pegasus.catalog.workflow.url"
PROP_CATALOG_WORKFLOW_TIMEOUT = "pegasus.catalog.workflow.timeout"
PROP_CATALOG_WORKFLOW_DB_TIMEOUT = "pegasus.catalog.workflow.db.timeout"
PROP_DASHBOARD_OUTPUT = "pegasus.dashboard.output"
PROP_MONITORD_OUTPUT = "pegasus.monitord.output"

CONNECTION_PROPERTIES = [
    PROP_CATALOG_MASTER_URL,
    PROP_CATALOG_REPLICA_DB_URL,
    PROP_CATALOG_WORKFLOW_URL,
    PROP_DASHBOARD_OUTPUT,
    PROP_MONITORD_OUTPUT,
]


# -------------------------------------------------------------------


class ConnectionError(Exception):
    def __init__(self, message, given_version=None, db_type=None):
        """
        :param message: Exception message
        :param given_version: Provided pegasus version
        :param db_type: Database type (JDBCRC, MASTER, WORKFLOW)
        """
        super().__init__(message)

        self.given_version = given_version
        self.db_type = db_type


class DBType:
    JDBCRC = "JDBCRC"
    MASTER = "MASTER"
    WORKFLOW = "WORKFLOW"


class DBKey:
    TIMEOUT = "timeout"


def connect(
    dburi,
    echo=False,
    schema_check=True,
    create=False,
    pegasus_version=None,
    force=False,
    props=None,
    db_type=None,
    backup=False,
    connect_args=None,
    verbose=True,
    print_version=True,
):
    """
    Connect to the provided URL database.
    :param dburi: DB URI
    :param echo:
    :param schema_check: Whether the schema should be checked
    :param create: Whether to create/update the database
    :param pegasus_version: Pegasus given version to create/check the database
    :param force:
    :param props:
    :param db_type: DB type (JDBCRC, MASTER, or WORKFLOW)
    :param backup: Whether to backup the database
    :param connect_args:
    :param verbose: Whether information log should be printed
    :param print_version: Whether to print the database version
    :return:
    """
    dburi = _parse_jdbc_uri(dburi)
    _validate(dburi)

    mask = _backup_db(dburi) if backup else None
    init = _db_is_file(dburi)

    try:
        # parse connection properties
        # PM-898 monitord sends props as None and connect_args has the
        # right values for the connection
        connect_args = _parse_props(dburi, props, db_type, connect_args)
        log.debug(
            "Connecting to: {} with connection params as {}".format(dburi, connect_args)
        )

        engine = create_engine(
            dburi, echo=echo, pool_recycle=True, connect_args=connect_args
        )
        engine.connect()
        if backup or not init:
            _check_db_permissions(dburi, db_type, mask)

    except exc.OperationalError as e:
        if "mysql" in dburi and "unknown database" in str(e).lower():
            raise ConnectionError(
                "MySQL database should be previously created: {} ({})".format(e, dburi),
                given_version=pegasus_version,
                db_type=db_type,
            )
        raise ConnectionError(
            "{} ({})".format(e, dburi), given_version=pegasus_version, db_type=db_type
        )
    except Exception as e:
        raise ConnectionError(
            "{} ({})".format(e, dburi), given_version=pegasus_version, db_type=db_type
        )

    Session = orm.sessionmaker(
        bind=engine, autoflush=False, autocommit=False, expire_on_commit=False
    )
    db = orm.scoped_session(Session)

    # Database creation
    if create:
        try:
            from Pegasus.db.admin.admin_loader import db_create

            db_create(
                dburi,
                engine,
                db,
                pegasus_version=pegasus_version,
                force=force,
                verbose=verbose,
            )

        except exc.OperationalError as e:
            if "database is locked" in str(e).lower():
                # database is locked, getting PIDs and commands
                p = urlparse(dburi)
                out, err = subprocess.Popen(
                    "fuser -u %s" % p.path,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                    cwd=os.getcwd(),
                ).communicate()

                pids = out.decode("utf8").strip().replace(" ", ",")
                out, err = subprocess.Popen(
                    'ps -o "pid user command" -p %s' % pids,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                    cwd=os.getcwd(),
                ).communicate()
                if err:
                    raise ConnectionError(
                        "Database is locked (%s): Unable to run ps command: %s"
                        % (dburi, err),
                        given_version=pegasus_version,
                        db_type=db_type,
                    )

                output = out.decode("utf8").strip()
                output = output[output.find("\n") + 1 :]
                raise ConnectionError(
                    "Database is locked ({}):\n{}".format(dburi, output),
                    given_version=pegasus_version,
                    db_type=db_type,
                )
            else:
                raise ConnectionError(
                    "{} ({})".format(e, dburi),
                    given_version=pegasus_version,
                    db_type=db_type,
                )

    if not create and schema_check:
        try:
            from Pegasus.db.admin.admin_loader import DBAdminError, db_verify

            db = orm.scoped_session(Session)
            db_verify(
                db,
                check=True,
                pegasus_version=pegasus_version,
                force=force,
                print_version=print_version,
            )
        except DBAdminError as e:
            e.db_type = db_type
            raise (e)

    return db


def connect_by_submitdir(
    submit_dir,
    db_type,
    config_properties=None,
    echo=False,
    schema_check=True,
    create=False,
    pegasus_version=None,
    force=False,
    cl_properties=None,
    print_version=True,
):
    """ Connect to the database from submit directory and database type """
    dburi = url_by_submitdir(
        submit_dir, db_type, config_properties, cl_properties=cl_properties
    )
    return connect(
        dburi,
        echo,
        schema_check,
        create=create,
        pegasus_version=pegasus_version,
        force=force,
        print_version=print_version,
    )


def connect_by_properties(
    config_properties,
    db_type,
    cl_properties=None,
    echo=False,
    schema_check=True,
    create=False,
    pegasus_version=None,
    force=False,
    verbose=True,
    print_version=True,
):
    """ Connect to the database from properties file and database type """
    props = properties.Properties()
    props.new(config_file=config_properties)
    _merge_properties(props, cl_properties)

    dburi = url_by_properties(config_properties, db_type, props=props)
    return connect(
        dburi,
        echo,
        schema_check,
        create=create,
        pegasus_version=pegasus_version,
        force=force,
        db_type=db_type,
        props=props,
        verbose=verbose,
        print_version=print_version,
    )


def url_by_submitdir(
    submit_dir, db_type, config_properties=None, top_dir=None, cl_properties=None
):
    """ Get URL from the submit directory """
    if not submit_dir:
        raise ConnectionError(
            "A submit directory should be provided with the type parameter.",
            db_type=db_type,
        )
    if not db_type:
        raise ConnectionError(
            "A type should be provided with the property file.", db_type=db_type
        )

    # From the submit dir, we need the wf_uuid
    # Getting values from the submit_dir braindump file
    top_level_wf_params = _parse_top_level_wf_params(submit_dir, top_dir)

    # Get the location of the properties file from braindump
    top_level_prop_file = None

    # Get properties tag from braindump
    if "properties" in top_level_wf_params:
        top_level_prop_file = top_level_wf_params["properties"]
        # Create the full path by using the submit_dir key from braindump
        if "submit_dir" in top_level_wf_params:
            top_level_prop_file = os.path.join(
                top_level_wf_params["submit_dir"], top_level_prop_file
            )

    return url_by_properties(
        config_properties,
        db_type,
        submit_dir,
        top_dir=top_dir,
        rundir_properties=top_level_prop_file,
        cl_properties=cl_properties,
    )


def url_by_properties(
    config_properties,
    db_type,
    submit_dir=None,
    top_dir=None,
    rundir_properties=None,
    cl_properties=None,
    props=None,
):
    """ Get URL from the property file """
    # Validate parameters
    if not db_type:
        raise ConnectionError(
            "A type should be provided with the property file.", db_type=db_type
        )

    # Parse, and process properties
    if not props:
        props = properties.Properties()
        props.new(config_file=config_properties, rundir_propfile=rundir_properties)
        _merge_properties(props, cl_properties)

    if db_type.upper() == DBType.JDBCRC:
        dburi = _get_jdbcrc_uri(props)
    elif db_type.upper() == DBType.MASTER:
        dburi = _get_master_uri(props)
    elif db_type.upper() == DBType.WORKFLOW:
        dburi = _get_workflow_uri(props, submit_dir, top_dir)
    else:
        raise ConnectionError("Invalid database type '%s'." % db_type, db_type=db_type)

    if dburi:
        log.debug("Using database: %s" % dburi)
        return dburi

    raise ConnectionError("Unable to find a database URI to connect.", db_type=db_type)


def get_wf_uuid(submit_dir):
    # From the submit dir, we need the wf_uuid
    # Getting values from the submit_dir braindump file
    top_level_wf_params = utils.slurp_braindb(submit_dir)

    # Return if we cannot parse the braindump.txt file
    if not top_level_wf_params:
        log.error("Unable to process braindump.txt in %s" % (submit_dir))
        return None

    # Get wf_uuid for this workflow
    wf_uuid = None
    if "wf_uuid" in top_level_wf_params:
        wf_uuid = top_level_wf_params["wf_uuid"]
    else:
        log.error("workflow id cannot be found in the braindump.txt ")
        return None

    return wf_uuid


def connect_to_master_db(user=None):
    """Connect to 'user's master database"""

    if user is None:
        user = getpass.getuser()

    u = users.get_user_by_username(user)

    dburi = u.get_master_db_url()

    return connect(dburi)


# -------------------------------------------------------------------

# This turns on foreign keys for SQLite3 connections
@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(conn, record):
    if isinstance(conn, SQLite3Connection):
        log.debug("Turning on foreign keys")
        cursor = conn.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()


def _merge_properties(props, cl_properties):
    if cl_properties:
        for property in cl_properties:
            if "=" not in property:
                raise ConnectionError("Malformed property: %s" % property)
            key, value = property.split("=")
            props.property(key, val=value)


def _get_jdbcrc_uri(props=None):
    """ Get JDBCRC URI from properties """
    if props:
        replica_catalog = props.property("pegasus.catalog.replica")
        if not replica_catalog:
            raise ConnectionError("'pegasus.catalog.replica' property not set.")

        if replica_catalog.upper() != DBType.JDBCRC:
            return None

        rc_info = {
            "driver": props.property("pegasus.catalog.replica.db.driver"),
            "url": props.property(PROP_CATALOG_REPLICA_DB_URL),
            "user": props.property("pegasus.catalog.replica.db.user"),
            "password": props.property("pegasus.catalog.replica.db.password"),
        }

        url = rc_info["url"]
        if not url:
            raise ConnectionError(
                "'%s' property not set." % PROP_CATALOG_REPLICA_DB_URL
            )
        url = _parse_jdbc_uri(url)
        o = urlparse(url)
        host = o.netloc
        database = o.path.replace("/", "")

        driver = rc_info["driver"]
        if not driver:
            raise ConnectionError(
                "'pegasus.catalog.replica.db.driver' property not set."
            )

        if driver.lower() == "mysql":
            return (
                "mysql://"
                + rc_info["user"]
                + ":"
                + rc_info["password"]
                + "@"
                + host
                + "/"
                + database
            )

        if driver.lower() == "sqlite":
            if "sqlite:" in url:
                return url
            connString = os.path.join(host, "workflow.db")
            return "sqlite:///" + connString

        if driver.lower() == "postgresql":
            return (
                "postgresql://"
                + rc_info["user"]
                + ":"
                + rc_info["password"]
                + "@"
                + host
                + "/"
                + database
            )

        log.debug("Invalid JDBCRC driver: %s" % rc_info["driver"])
    return None


def _get_master_uri(props=None):
    """ Get MASTER URI """
    if props:
        dburi = props.property(PROP_CATALOG_MASTER_URL)
        if dburi:
            return dburi
        dburi = props.property(PROP_DASHBOARD_OUTPUT)
        if dburi:
            return dburi

    homedir = os.getenv("HOME", None)
    if homedir is None:
        raise ConnectionError(
            "Environment variable HOME not defined, set %s property to point to the Dashboard database."
            % PROP_DASHBOARD_OUTPUT
        )

    dir = os.path.join(homedir, ".pegasus")

    # check for writability and create directory if required
    if not os.path.isdir(dir):
        try:
            os.mkdir(dir, 0o755)
        except OSError:
            raise ConnectionError("Unable to create directory: %s" % dir)
    elif not os.access(dir, os.W_OK):
        log.warning("Unable to write to directory: %s" % dir)
        return None

    # directory exists, touch the file and set permissions
    filename = os.path.join(dir, "workflow.db")
    if not os.access(filename, os.F_OK):
        try:
            # touch the file
            open(filename, "w").close()
            os.chmod(filename, 0o600)
        except Exception as e:
            log.warning("unable to initialize MASTER db %s." % filename)
            log.exception(e)
            return None
    elif not os.access(filename, os.W_OK):
        log.warning("No read access for file: %s" % filename)
        return None

    return "sqlite:///" + filename


def _get_workflow_uri(props=None, submit_dir=None, top_dir=None):
    """ Get WORKFLOW URI """
    if props:
        dburi = props.property(PROP_CATALOG_WORKFLOW_URL)
        if dburi:
            return dburi
        dburi = props.property(PROP_MONITORD_OUTPUT)
        if dburi:
            return dburi

    top_level_wf_params = None
    if submit_dir or top_dir:
        # From the submit dir, we need the wf_uuid
        # Getting values from the submit_dir braindump file
        top_level_wf_params = _parse_top_level_wf_params(submit_dir, top_dir)

    if not top_level_wf_params:
        return None

    # The default case is a .stampede.db file with the dag name as base
    dag_file_name = ""
    if "dag" in top_level_wf_params:
        dag_file_name = top_level_wf_params["dag"]
    else:
        raise ConnectionError("DAG file name cannot be found in the braindump.txt.")

    # Create the sqllite db url
    dag_file_name = os.path.basename(dag_file_name)
    output_db_file = os.path.join(
        top_level_wf_params["submit_dir"],
        dag_file_name[: dag_file_name.find(".dag")] + ".stampede.db",
    )

    # if path does not exist, fallback on the submit_dir if provided
    if not os.path.exists(output_db_file) and submit_dir:
        output_db_file = os.path.join(
            submit_dir, dag_file_name[: dag_file_name.find(".dag")] + ".stampede.db"
        )

    dburi = "sqlite:///" + output_db_file
    return dburi


def _parse_jdbc_uri(dburi):
    if dburi:
        if dburi.startswith("jdbc:"):
            dburi = dburi.replace("jdbc:", "")
            if dburi.startswith("sqlite:"):
                dburi = dburi.replace("sqlite:", "sqlite:///")
    return dburi


def _validate(dburi):
    try:
        if dburi:
            if dburi.startswith("postgresql:"):
                import psycopg2  # noqa: F401
            if dburi.startswith("mysql:"):
                import MySQLdb  # noqa: F401

    except ImportError as e:
        raise ConnectionError("Missing Python module: {} ({})".format(e, dburi))


def _check_db_permissions(dburi, db_type, mask=None):
    if urlparse(dburi).scheme == "sqlite" and dburi.lower() != "sqlite://":
        path = urlparse(dburi).path[1:]
        if db_type:
            if db_type == DBType.MASTER or db_type == DBType.WORKFLOW:
                os.chmod(path, 0o744)
        elif path == os.path.expanduser("~") + "/.pegasus/workflow.db":
            os.chmod(path, 0o744)

        if mask:
            os.chmod(path, mask)


def _parse_props(dburi, props, db_type=None, connect_args=None):
    """

    :param db:
    :param props:
    :param db_type:
    :return:
    """
    if not connect_args:
        connect_args = {}

    if props and dburi.lower().startswith("sqlite") and db_type:
        if DBKey.TIMEOUT not in connect_args:
            try:
                timeout = None
                if db_type == DBType.MASTER:
                    timeout = _get_timeout_property(
                        props,
                        PROP_CATALOG_MASTER_TIMEOUT,
                        PROP_CATALOG_MASTER_DB_TIMEOUT,
                    )
                elif db_type == DBType.WORKFLOW:
                    timeout = _get_timeout_property(
                        props,
                        PROP_CATALOG_WORKFLOW_TIMEOUT,
                        PROP_CATALOG_WORKFLOW_DB_TIMEOUT,
                    )
                elif db_type == DBType.JDBCRC:
                    timeout = _get_timeout_property(
                        props,
                        PROP_CATALOG_REPLICA_TIMEOUT,
                        PROP_CATALOG_REPLICA_DB_TIMEOUT,
                    )
                if not timeout:
                    timeout = _get_timeout_property(
                        props, PROP_CATALOG_ALL_TIMEOUT, PROP_CATALOG_ALL_DB_TIMEOUT
                    )

                if timeout:
                    connect_args[DBKey.TIMEOUT] = timeout

            except ValueError as e:
                raise ConnectionError(
                    "Timeout properties should be set in seconds: {} ({})".format(
                        e, dburi
                    ),
                    db_type=db_type,
                )

    return connect_args


def _get_timeout_property(props, prop_name1, prop_name2):
    """

    :param props:
    :param prop_name1:
    :param prop_name2:
    :return:
    """
    if props.property(prop_name1):
        return int(props.property(prop_name1))

    elif props.property(prop_name2):
        return int(props.property(prop_name2))

    return None


def _parse_top_level_wf_params(submit_dir, top_dir):
    """
    Parse the top level workflow parameters.
    :param submit_dir: path of the directory
    :param top_dir: path to the workflow top directory
    :return: top level workflow parameters
    """
    dir = submit_dir
    if top_dir:
        dir = top_dir

    top_level_wf_params = utils.slurp_braindb(dir)

    # Return if we cannot parse the braindump.txt file
    if not top_level_wf_params:
        raise ConnectionError("File 'braindump.txt' not found in %s." % dir)

    if top_level_wf_params["root_wf_uuid"] == top_level_wf_params["wf_uuid"]:
        return top_level_wf_params

    while True:
        dir = os.path.abspath(os.path.join(os.path.abspath(dir), ".."))
        if dir == os.path.realpath("/.."):
            top_level_wf_params = None
            break

        top_level_wf_params = utils.slurp_braindb(dir)
        if (
            top_level_wf_params
            and top_level_wf_params["root_wf_uuid"] == top_level_wf_params["wf_uuid"]
        ):
            break

    if not top_level_wf_params:
        raise ConnectionError("Unable to find file 'braindump.txt' in parent folders.")

    return top_level_wf_params


def _backup_db(dburi):
    """
    Create a copy of the database (SQLite).
    :param dburi: DB URI
    """
    mask = None
    if urlparse(dburi).scheme == "sqlite":
        db_path = urlparse(dburi).path[1:]
        if os.path.isfile(db_path):
            log.info("Rotating sqlite db file %s" % db_path)
            mask = os.stat(db_path)[ST_MODE]
            utils.rotate_log_file(db_path)

    return mask


def _db_is_file(dburi):
    if urlparse(dburi).scheme == "sqlite":
        db_path = urlparse(dburi).path[1:]
        return os.path.isfile(db_path)
    return True



import ast
import logging


class InvalidSortError(Exception):
    pass


def sort_parse(clause):
    """
    Parses sort clause.

    Sort Clause Identifier Rules:
        Prefix + Identifier

    Prefix Rules:
        Is optional
        Must start with a-z OR A-Z
        Can contain a-z OR A-Z OR 0-9 OR _
        Must end with .

    Identifier Rules:
        Is mandatory
        Must start with a-z OR A-Z
        Can contain a-z OR A-Z OR 0-9 OR _

    :param clause: The clause to be parsed.
    :type clause: str

    :raises InvalidSortError: Raised if the `clause` is not valid.

    :returns: list containing tuples of field name, and sort order.
    :rtype: list

    .. example::

        >>> sort_parse("-prefix.identifier")
        [('prefix', 'identifier', 'DESC')]

        >>> sort_parse("a, +a.b")
        [('a', 'ASC'), ('a', 'b', 'ASC')]
    """
    try:
        n = ast.parse(clause.strip(",") + ",", mode="eval").body
    except SyntaxError as e:
        raise InvalidSortError("Invalid sort: %s" % e)

    if not isinstance(n, ast.Tuple):
        raise InvalidSortError("Invalid condition: must evaluate to a boolean value")

    return _SortEvaluator().visit(n)


class _SortEvaluator(ast.NodeVisitor):
    def __init__(self):
        self._sort_dir = None
        self.reset_context()
        self._log = logging.getLogger(__name__)

    def reset_context(self):
        self._sort_dir = "ASC"

    def visit_Tuple(self, n):
        sort = []
        for elt in n.elts:
            self.reset_context()
            sort.append(tuple(self.visit(elt)) + (self._sort_dir,))
        return sort

    def visit_UnaryOp(self, n):
        self.visit(n.op)
        return self.visit(n.operand)

    def visit_UAdd(self, n):
        pass

    def visit_USub(self, n):
        self._sort_dir = "DESC"

    def visit_Attribute(self, n):
        self._log.info("Attribute <%s> <%s>", n.value, n.attr)
        value = self.visit(n.value)
        value.append(n.attr)
        return value

    def visit_Name(self, n):
        return [n.id]

    def generic_visit(self, n):
        raise InvalidSortError(
            "Invalid sort clause at Line <%d>, Col <%d>" % (n.lineno, n.col_offset)
        )


def _main():
    import sys

    logging.basicConfig(level=logging.DEBUG)
    logging.debug("Expression <%s>", sys.argv[1])
    result = sort_parse(sys.argv[1])
    logging.debug("Evaluation Result <%s>", result)


if __name__ == "__main__":
    _main()



import hashlib
import logging

from flask import g, make_response, request

from Pegasus.service import cache
from Pegasus.service._serialize import jsonify
from Pegasus.service.base import OrderedDict
from Pegasus.service.monitoring import monitoring as blueprint
from Pegasus.service.monitoring.queries import (
    MasterWorkflowQueries,
    StampedeWorkflowQueries,
)

log = logging.getLogger(__name__)

JSON_HEADER = {"Content-Type": "application/json"}


@blueprint.url_value_preprocessor
def pull_m_wf_id(endpoint, values):
    """
    If the requested endpoint contains a value for m_wf_id variable then extract it and set it in g.m_wf_id.
    """
    if values and "m_wf_id" in values:
        g.m_wf_id = values["m_wf_id"]


@blueprint.url_value_preprocessor
def pull_url_context(endpoint, values):
    """
    Create a context which can be used when generating url in link section of the responses.
    """
    url_context = {}
    keys = ["wf_id", "job_id", "task_id", "job_instance_id", "host_id", "instance_id"]

    if values:
        for key in keys:
            if key in values:
                url_context[key] = values[key]

        else:
            if url_context:
                g.url_context = url_context


@blueprint.before_request
def compute_stampede_db_url():
    """
    If the requested endpoint requires connecting to a STAMPEDE database, then determine STAMPEDE DB URL and store it
    in g.stampede_db_url. Also, set g.m_wf_id to be the root workflow's uuid
    """
    if "/workflow" not in request.path or "m_wf_id" not in g:
        return

    md5sum = hashlib.md5()
    md5sum.update(g.master_db_url.encode("utf-8"))
    m_wf_id = g.m_wf_id

    def _get_cache_key(key_suffix):
        return "{}.{}".format(md5sum.hexdigest(), key_suffix)

    cache_key = _get_cache_key(m_wf_id)

    if cache.get(cache_key):
        log.debug("Cache Hit: compute_stampede_db_url %s" % cache_key)
        root_workflow = cache.get(cache_key)

    else:
        log.debug("Cache Miss: compute_stampede_db_url %s" % cache_key)
        queries = MasterWorkflowQueries(g.master_db_url)
        root_workflow = queries.get_root_workflow(m_wf_id)
        queries.close()

        cache.set(_get_cache_key(root_workflow.wf_id), root_workflow, timeout=600)
        cache.set(_get_cache_key(root_workflow.wf_uuid), root_workflow, timeout=600)

    g.url_m_wf_id = root_workflow.wf_id
    g.m_wf_id = root_workflow.wf_uuid
    g.stampede_db_url = root_workflow.db_url


@blueprint.before_request
def get_query_args():
    g.query_args = {}

    def to_int(q_arg, value):
        try:
            return int(value)
        except ValueError as e:
            log.exception(
                "Query Argument {} = {} is not a valid int".format(q_arg, value)
            )
            e = ValueError(
                "Expecting integer for argument {}, found {!r}".format(
                    q_arg, str(value)
                )
            )
            e.codes = ("INVALID_QUERY_ARGUMENT", 400)
            raise e from None

    def to_str(q_arg, value):
        return value

    def to_bool(q_arg, value):
        value = value.strip().lower()

        if value in {"1", "true"}:
            return True

        elif value in {"0", "false"}:
            return False

        else:
            log.exception(
                "Query Argument {} = {} is not a valid boolean".format(q_arg, value)
            )
            e = ValueError(
                "Expecting boolean for argument {}, found {!r}".format(
                    q_arg, str(value)
                )
            )
            e.codes = ("INVALID_QUERY_ARGUMENT", 400)
            raise e

    query_args = OrderedDict(
        [
            ("pretty-print", to_bool),
            ("start-index", to_int),
            ("max-results", to_int),
            ("query", to_str),
            ("order", to_str),
        ]
    )

    for arg, cast in query_args.items():
        if arg in request.args:
            g.query_args[arg.replace("-", "_")] = cast(arg, request.args.get(arg))


"""
Root Workflow

{
    "wf_id"             : int:wf_id,
    "wf_uuid"           : string:wf_uuid,
    "submit_hostname"   : string:submit_hostname,
    "submit_dir"        : string:submit_dir,
    "planner_arguments" : string:planner_arguments,
    "planner_version"   : string:planner_version,
    "user"              : string:user,
    "grid_dn"           : string:grid_dn,
    "dax_label"         : string:dax_label,
    "dax_version"       : string:dax_version,
    "dax_file"          : string:dax_file,
    "dag_file_name"     : string:dag_file_name,
    "timestamp"         : int:timestamp,
    "workflow_state"    : object:workflow_state,
    "_links"            : {
        "workflow" : href:workflow
    }
}
"""


@blueprint.route("/root")
def get_root_workflows(username):
    """
    Returns a collection of root level workflows.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response.

    :statuscode 200: OK
    :statuscode 204: No content; when no workflows found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Root Workflow
    """
    queries = MasterWorkflowQueries(g.master_db_url)
    paged_response = queries.get_root_workflows(**g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>")
def get_root_workflow(username, m_wf_id):
    """
    Returns root level workflow identified by m_wf_id.

    :query boolean pretty-print: Return formatted JSON response.

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: Root Workflow
    """
    queries = MasterWorkflowQueries(g.master_db_url)
    record = queries.get_root_workflow(m_wf_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Workflow

{
    "wf_id"             : int:wf_id,
    "root_wf_id"        : int:root_wf_id,
    "parent_wf_id"      : int:parent_wf_id,
    "wf_uuid"           : string:wf_uuid,
    "submit_hostname"   : string:submit_hostname,
    "submit_dir"        : string:submit_dir,
    "planner_arguments" : string:planner_arguments,
    "planner_version"   : string:planner_version,
    "user"              : string:user,
    "grid_dn"           : string:grid_dn,
    "dax_label"         : string:dax_label,
    "dax_version"       : string:dax_version,
    "dax_file"          : string:dax_file,
    "dag_file_name"     : string:dag_file_name,
    "timestamp"         : int:timestamp,
    "_links"            : {
        "workflow_meta"  : href:workflow_meta,
        "workflow_state" : href:workflow_state,
        "job"            : href:job,
        "task"           : href:task,
        "host"           : href:host,
        "invocation"     : href:invocation
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow")
def get_workflows(username, m_wf_id):
    """
    Returns a collection of workflows.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no workflows found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Workflow
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflows(g.m_wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>")
def get_workflow(username, m_wf_id, wf_id):
    """
    Returns workflow identified by m_wf_id, wf_id.

    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: Workflow
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    record = queries.get_workflow(wf_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Workflow Meta

{
    "key"         : string:key,
    "value"       : string:value,
    "_links"      : {
        "workflow" : <href:workflow>
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/meta")
def get_workflow_meta(username, m_wf_id, wf_id):
    """
    Returns a collection of workflow's metadata.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no workflow metadata found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: WorkflowMeta
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_meta(g.m_wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


"""
Workflow Files

{
    "wf_id"  : int:wf_id,
    "lfn_id" : string:lfn_id,
    "lfn"    : string:lfn,
    "pfns"   : [
         {
            "pfn_id" : <int:pfn_id>
            "pfn"    : <string:pfn>
            "site"   : <string:site>
         }
    ],
    "meta" : [
         {
            "meta_id" : <int:meta_id>
            "key"     : <string:key>
            "value"   : <string:value>
         }
    ],
    "_links"      : {
        "workflow" : <href:workflow>
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/files")
def get_workflow_files(username, m_wf_id, wf_id):
    """
    Returns a collection of workflows.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no workflows found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Workflow
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_files(g.m_wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


"""
Workflow State

{
    "wf_id"         : int:wf_id,
    "state"         : string:state,
    "status"        : int:status,
    "restart_count" : int:restart_count,
    "timestamp"     : datetime:timestamp,
    "_links"        : {
        "workflow" : <href:workflow>
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/state")
def get_workflow_state(username, m_wf_id, wf_id):
    """
    Returns a collection of Workflow States.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no workflowstates found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: WorkflowState
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_state(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


"""
Job

{
    "job_id"      : int: job_id,
    "exec_job_id" : string: exec_job_id,
    "submit_file" : string: submit_file,
    "type_desc"   : string: type_desc,
    "max_retries" : int: max_retries,
    "clustered"   : bool: clustered,
    "task_count"  : int: task_count,
    "executable"  : string: executable,
    "argv"        : string: argv,
    "task_count"  : int:task_count,
    "_links"      : {
        "workflow"     : href:workflow,
        "task"         : href:task,
        "job_instance" : href:job_instance
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job")
def get_workflow_jobs(username, m_wf_id, wf_id):
    """
    Returns a collection of Jobs.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no jobs found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Job
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_jobs(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job/<int:job_id>")
def get_job(username, m_wf_id, wf_id, job_id):
    """
    Returns job identified by m_wf_id, wf_id, job_id.

    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: Job
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    record = queries.get_job(job_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Host

{
    "host_id"      : int:host_id,
    "site"         : string:site,
    "hostname"     : string:hostname,
    "ip"           : string:ip,
    "uname"        : string:uname,
    "total_memory" : string:total_memory,
    "_links"       : {
        "workflow" : href:workflow
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/host")
def get_workflow_hosts(username, m_wf_id, wf_id):
    """
    Returns a collection of Hosts.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no hosts found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Host
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_hosts(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/host/<int:host_id>")
def get_host(username, m_wf_id, wf_id, host_id):
    """
    Returns host identified by m_wf_id, wf_id, host_id.

    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: Host
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    record = queries.get_host(host_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Job State

{
    "job_instance_id"     : "<int:job_instance_id>",
    "state"               : "<string:state>",
    "jobstate_submit_seq" : "<int:jobstate_submit_seq>",
    "timestamp"           : "<int:timestamp>",
    "_links"              : {
        "job_instance" : "href:job_instance"
    }
}
"""


@blueprint.route(
    "/root/<string:m_wf_id>/workflow/<string:wf_id>/job/<int:job_id>/job-instance/<int:job_instance_id>/state"
)
def get_job_instance_states(username, m_wf_id, wf_id, job_id, job_instance_id):
    """
    Returns a collection of Job States.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no jobstates found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: JobState
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_job_instance_states(
        wf_id, job_id, job_instance_id, **g.query_args
    )

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


"""
Task

{
    "task_id"        : int:task_id,
    "abs_task_id"    : string:abs_task_id,
    "type_desc"      : string: type_desc,
    "transformation" : string:transformation,
    "argv"           : string:argv,
    "task_count"     : int:task_count,
    "_links"         : {
        "workflow"  : href:workflow,
        "job"       : href:job,
        "task_meta" : href:task_meta
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/task")
def get_workflow_tasks(username, m_wf_id, wf_id):
    """
    Returns a collection of Tasks.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no tasks found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Task
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_tasks(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job/<int:job_id>/task")
def get_job_tasks(username, m_wf_id, wf_id, job_id):
    """
    Returns a collection of Tasks.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no tasks found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Task
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_job_tasks(wf_id, job_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/task/<int:task_id>")
def get_task(username, m_wf_id, wf_id, task_id):
    """
    Returns task identified by m_wf_id, wf_id, task_id.

    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: Task
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    record = queries.get_task(task_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Task Meta

{
    "key"         : string:key,
    "value"       : string:value,
    "_links"      : {
        "task" : "<href:task>"
    }
}
"""


@blueprint.route(
    "/root/<string:m_wf_id>/workflow/<string:wf_id>/task/<int:task_id>/meta"
)
def get_task_meta(username, m_wf_id, wf_id, task_id):
    """
    Returns a collection of task's metadata.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no workflow metadata found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: TaskMeta
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_task_meta(task_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


"""
Job Instance

{
    "job_instance_id"   : int:job_instance_id,
    "host_id"           : int:host_id,
    "job_submit_seq"    : int:job_submit_seq,
    "sched_id"          : string:sched_id,
    "site"              : string:site,
    "user"              : string:user,
    "work_dir"          : string:work_dir,
    "cluster_start"     : int:cluster_start,
    "cluster_duration"  : int:cluster_duration,
    "local_duration"    : int:local_duration,
    "subwf_id"          : int:subwf_id,
    "stdout_text"       : string:stdout_text,
    "stderr_text"       : string:stderr_text,
    "stdin_file"        : string:stdin_file,
    "stdout_file"       : string:stdout_file,
    "stderr_file"       : string:stderr_file,
    "multiplier_factor" : int:multiplier_factor,
    "exitcode"          : int:exitcode,
    "_links"            : {
        "job_state"  : href:job_state,
        "host"       : href:host,
        "invocation" : href:invocation,
        "job"        : href:job
    }
}
"""


@blueprint.route(
    "/root/<string:m_wf_id>/workflow/<string:wf_id>/job/<int:job_id>/job-instance"
)
def get_job_instances(username, m_wf_id, wf_id, job_id):
    """
    Returns a collection of JobInstances.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no job instances found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: JobInstance
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_job_instances(
        wf_id,
        job_id,
        recent=request.args.get("recent", "false") == "true",
        **g.query_args,
    )

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route(
    "/root/<string:m_wf_id>/workflow/<string:wf_id>/job-instance/<int:job_instance_id>"
)
def get_job_instance(username, m_wf_id, wf_id, job_instance_id):
    """
    Returns job instance identified by m_wf_id, wf_id, job_id, job_instance_id.

    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: JobInstance
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    record = queries.get_job_instance(job_instance_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Invocation

{
    "invocation_id"   : int:invocation_id,
    "job_instance_id" : int:job_instance_id,
    "abs_task_id"     : string:abs_task_id,
    "task_submit_seq" : int:task_submit_seq,
    "start_time"      : int:start_time,
    "remote_duration" : int:remote_duration,
    "remote_cpu_time" : int:remote_cpu_time,
    "exitcode"        : int:exitcode,
    "transformation"  : string:transformation,
    "executable"      : string:executable,
    "argv"            : string:argv,
    "_links"          : {
        "workflow"     : href:workflow,
        "job_instance" : href:job_instance
    }
}
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/invocation")
def get_workflow_invocations(username, m_wf_id, wf_id):
    """
    Returns a collection of Invocations.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no invocations found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Invocation
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_workflow_invocations(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route(
    "/root/<string:m_wf_id>/workflow/<string:wf_id>/job/<int:job_id>/job-instance/<int:job_instance_id>/invocation"
)
def get_job_instance_invocations(username, m_wf_id, wf_id, job_id, job_instance_id):
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_job_instance_invocations(
        wf_id, job_id, job_instance_id, **g.query_args
    )

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route(
    "/root/<string:m_wf_id>/workflow/<string:wf_id>/invocation/<int:invocation_id>"
)
def get_invocation(username, m_wf_id, wf_id, invocation_id):
    """
    Returns invocation identified by m_wf_id, wf_id, invocation_id.

    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure
    :statuscode 404: Not found

    :return type: Record
    :return resource: Invocation
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    record = queries.get_invocation(invocation_id)

    #
    # Generate JSON Response
    #
    response_json = jsonify(record)

    return make_response(response_json, 200, JSON_HEADER)


"""
Utilities
"""


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job/running")
def get_running_jobs(username, m_wf_id, wf_id):
    """
    Returns a collection of running Jobs.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no jobs found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Job
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_running_jobs(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job/successful")
def get_successful_jobs(username, m_wf_id, wf_id):
    """
    Returns a collection of successful Jobs.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no jobs found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Job
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_successful_jobs(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job/failed")
def get_failed_jobs(username, m_wf_id, wf_id):
    """
    Returns a collection of failed Jobs.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no jobs found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Job
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_failed_jobs(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)


@blueprint.route("/root/<string:m_wf_id>/workflow/<string:wf_id>/job/failing")
def get_failing_jobs(username, m_wf_id, wf_id):
    """
    Returns a collection of failing Jobs.

    :query int start-index: Return results starting from record <start-index> (0 indexed)
    :query int max-results: Return a maximum of <max-results> records
    :query string query: Search criteria
    :query string order: Sorting criteria
    :query boolean pretty-print: Return formatted JSON response

    :statuscode 200: OK
    :statuscode 204: No content; when no jobs found.
    :statuscode 400: Bad request
    :statuscode 401: Authentication failure
    :statuscode 403: Authorization failure

    :return type: Collection
    :return resource: Job
    """
    queries = StampedeWorkflowQueries(g.stampede_db_url)

    paged_response = queries.get_failing_jobs(wf_id, **g.query_args)

    if paged_response.total_records == 0:
        log.debug("Total records is 0; returning HTTP 204 No content")
        return make_response("", 204, JSON_HEADER)

    #
    # Generate JSON Response
    #
    response_json = jsonify(paged_response)

    return make_response(response_json, 200, JSON_HEADER)



#!/usr/bin/env python3

import os
import re
import shutil
import subprocess
from pathlib import Path

import click

from Pegasus.tools.utils import slurp_braindb


@click.command()
@click.pass_context
@click.option(
    # TODO: Remove `--dagid` as it is replaced with `--dag-id`
    "--dag-id",
    "--dagid",
    "-d",
    metavar="<dag-id>",
    help="The id of the dag to be removed.",
)
@click.argument(
    "submit-dir",
    required=False,
    type=click.Path(file_okay=False, dir_okay=True, readable=True, exists=True),
)
def pegasus_remove(ctx, dag_id=None, verbose=False, submit_dir=None):
    """pegasus-remove helps you remove an entire workflow."""
    if not submit_dir and not dag_id:
        print("You must provide either a dag_id or dagdirectory to remove a workflow.")
        ctx.exit(1)

    if submit_dir:
        cwd = os.getcwd()

        submit_dir = str(Path(submit_dir).resolve())
        try:
            os.chdir(submit_dir)
        except PermissionError:
            click.secho(
                click.style("Error: ", fg="red", bold=True)
                + "Cannot change to directory %s" % submit_dir
            )
            ctx.exit(1)

        config = slurp_braindb(submit_dir)
        if not config:
            click.secho(
                click.style("Error: ", fg="red", bold=True)
                + "%s is not a valid submit-dir" % submit_dir
            )
            ctx.exit(1)

        dag_log_file = config["dag"] + ".dagman.out"
        pattern = re.compile(r"\.([0-9\.]+) \(CONDOR_DAGMAN\) STARTING UP")

        with open(dag_log_file) as fp:
            for line in fp.readlines():
                match = pattern.search(line)
                if match:
                    dag_id = match.group(1)
            else:
                if not dag_id:
                    click.secho(
                        click.style("Error: ", fg="red", bold=True)
                        + "You must provide either a dag-id or dag-directory to remove a workflow."
                    )
                    ctx.exit(1)

        os.chdir(cwd)

    if dag_id:
        condor_rm = shutil.which("condor_rm")
        cmd = (condor_rm, dag_id)

        rv = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if rv.returncode == 0:
            click.echo(rv.stdout.decode().strip())
            click.secho("✨ Success", fg="green")
        else:
            click.echo(rv.stderr.decode().strip())
            click.secho("Error ", fg="red", bold=True)
            ctx.exit(42)


if __name__ == "__main__":
    pegasus_remove()



#!/usr/bin/env python3

import colorsys
import os
import shutil
import subprocess
import sys
import xml.sax
import xml.sax.handler
from collections import OrderedDict
from functools import cmp_to_key
from optparse import OptionParser
from pathlib import Path
from tempfile import NamedTemporaryFile

from Pegasus import yaml

COLORS = [
    "#1b9e77",
    "#d95f02",
    "#7570b3",
    "#e7298a",
    "#66a61e",
    "#e6ab02",
    "#a6761d",
    "#666666",
    "#8dd3c7",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#d9d9d9",
    "#bc80bd",
    "#ccebc5",
    "#ffed6f",
    "#ffffb3",
    "#aadd88",
    "#889933",
    "#22bbcc",
    "#d9dbb5",
]


def rgb2hex(r, g, b):
    return "#{:02x}{:02x}{:02x}".format(int(r * 255), int(g * 255), int(b * 255))


# Generate some colors to add to the list
s = 0.7
for l in [0.70, 0.55]:
    for h in range(0, 101, 10):
        if h == 40:
            continue
        rgb = colorsys.hls_to_rgb(h / 100.0, l, s)
        COLORS.append(rgb2hex(*rgb))


class DAG:
    def __init__(self):
        self.nodes = OrderedDict()


class Node:
    def __init__(self, _id=None):
        self.id = _id
        self.label = None
        self.level = 0
        self.parents = []
        self.children = []
        self.mark = 0
        self.closure = set()

    def renderNode(self, renderer):
        pass

    def renderEdge(self, renderer, parent):
        renderer.renderEdge(parent.id, self.id)

    def __repr__(self):
        return "({}, {})".format(self.id, self.label)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)


class Job(Node):
    def __init__(self):
        Node.__init__(self)
        self.xform = None

    def renderNode(self, renderer):
        if renderer.label_type == "xform":
            label = self.xform
        elif renderer.label_type == "id":
            label = "%s" % self.id
        elif renderer.label_type == "xform-id":
            label = "{}\\n{}".format(self.xform, self.id)
        elif renderer.label_type == "label-xform":
            if len(self.label) > 0:
                label = "{}\\n{}".format(self.label, self.xform)
            else:
                label = self.xform
        elif renderer.label_type == "label-id":
            if len(self.label) > 0:
                label = "{}\\n{}".format(self.label, self.id)
            else:
                label = self.id
        else:
            label = self.label
        color = renderer.getcolor(self.xform)
        renderer.renderNode(self.id, label, color)


class File(Node):
    def __init__(self):
        Node.__init__(self)

    def renderNode(self, renderer):
        renderer.renderNode(self.id, self.label, fillcolor="#ffed6f", shape="rect")


class Stack:
    def __init__(self):
        self.items = []

    def push(self, item):
        self.items.append(item)

    def pop(self):
        return self.items.pop()

    def peek(self, index=1):
        index = 0 - index
        return self.items[index]


class DAXHandler(xml.sax.handler.ContentHandler):
    """
    This is a DAX file parser
    """

    def __init__(self, files):
        self.files = files
        self.elements = Stack()
        self.last_job = None
        self.dag = DAG()
        self.nodes = self.dag.nodes

    def startElement(self, name, attrs):
        self.elements.push(name)

        if name in ["job", "dax", "dag"]:
            job = Job()

            job.id = attrs.get("id")
            if job.id is None:
                raise Exception("Invalid DAX: attribute 'id' missing")

            if name == "job":
                job.xform = attrs.get("name")
                if job.xform is None:
                    raise Exception("Invalid DAX: job name missing for job %s" % job.id)
                ns = attrs.get("namespace")
                version = attrs.get("version")
                if ns is not None:
                    job.xform = ns + "::" + job.xform
                if version is not None:
                    job.xform = job.xform + ":" + version
            elif name == "dax":
                job.xform = "pegasus::dax"
            else:
                job.xform = "pegasus::dag"

            job.label = attrs.get("node-label")
            if job.label is None:
                job.label = attrs.get("file")
                if job.label is None:
                    job.label = job.xform

            self.nodes[job.id] = job
            self.last_job = job
        elif name == "uses":
            if not self.files:
                return

            # Ignore uses inside all elements except job
            if self.elements.peek(2) != "job":
                return

            namespace = attrs.get("namespace")
            version = attrs.get("version")
            filename = attrs.get("name") or attrs.get("file")

            if filename is None:
                raise Exception("name attribute missing on <uses>")

            if namespace is not None:
                filename = namespace + "::" + filename

            if version is not None:
                filename = filename + ":" + version

            if filename in self.nodes:
                f = self.nodes[filename]
            else:
                f = File()
                f.id = f.label = filename
                self.nodes[filename] = f

            link = attrs.get("link") or "input"
            link = link.lower()
            if link == "input":
                f.children.append(self.last_job)
                self.last_job.parents.append(f)
            elif link == "output":
                f.parents.append(self.last_job)
                self.last_job.children.append(f)
            elif link == "inout":
                print(
                    "WARNING: inout file {} of {} creates a cycle.".format(
                        f.id, self.last_job
                    )
                )
                f.children.append(self.last_job)
                f.parents.append(self.last_job)
                self.last_job.parents.append(f)
                self.last_job.children.append(f)
            elif link == "none":
                pass
            else:
                raise Exception("Unrecognized link value: %s" % link)
        elif name == "child":
            self.lastchild = attrs.get("ref")
        elif name == "parent":
            if self.lastchild is None:
                raise Exception("Invalid DAX: <parent> outside <child>")
            pid = attrs.get("ref")
            child = self.nodes[self.lastchild]
            parent = self.nodes[pid]
            child.parents.append(parent)
            parent.children.append(child)

    def endElement(self, name):
        self.elements.pop()
        if name == "child":
            self.lastchild = None


def parse_daxfile(fname, files=False):
    """
    Parse DAG from a Pegasus DAX file.
    """
    handler = DAXHandler(files)
    parser = xml.sax.make_parser()
    parser.setContentHandler(handler)
    f = open(fname)
    parser.parse(f)
    f.close()
    return handler.dag


def parse_xform_name(path):
    """
    Parse the transformation name from a submit script. Usually the
    transformation is in a special classad called '+pegasus_wf_xformation'.
    For special pegasus jobs (create_dir, etc.) set the name manually.
    """
    # Handle special cases
    fname = os.path.basename(path)
    if fname.startswith("create_dir_"):
        return "pegasus::create_dir"
    if fname.startswith("stage_in_"):
        return "pegasus::stage_in"
    if fname.startswith("stage_out_"):
        return "pegasus::stage_out"
    if fname.startswith("stage_inter_"):
        return "pegasus::stage_inter"
    if fname.startswith("stage_worker_"):
        return "pegasus::stage_worker"
    if fname.startswith("register_"):
        return "pegasus::register"
    if fname.startswith("clean_up_"):
        return "pegasus::clean_up"

    # Get it from the submit file
    if os.path.isfile(path):
        f = open(path)
        for line in f.readlines():
            if "+pegasus_wf_xformation" in line:
                return line.split('"')[1]

    # Otherwise, guess the xform by stripping digits from the name
    name = fname.replace(".sub", "")
    return "".join(c for c in name if not "0" <= c <= "9")


def parse_dagfile(fname):
    """
    Parse a DAG from a dagfile.
    """
    dagdir = os.path.dirname(fname)
    dag = DAG()
    jobs = dag.nodes
    f = open(fname)
    for line in f.readlines():
        line = line.strip()
        if line.startswith("JOB"):
            rec = line.split()
            job = Job()
            if len(rec) < 3:
                raise Exception("Invalid line:", line)
            job.id = rec[1]  # Job id
            subfile = rec[2]  # submit script
            if not os.path.isabs(subfile):
                subfile = os.path.join(dagdir, subfile)
            job.xform = parse_xform_name(subfile)
            job.label = job.id
            jobs[job.id] = job
        elif line.startswith("PARENT"):
            rec = line.split()
            if len(rec) < 4:
                raise Exception("Invalid line:", line)
            p = jobs[rec[1]]
            c = jobs[rec[3]]
            p.children.append(c)
            c.parents.append(p)
    f.close()

    return dag


def parse_yamlfile(fname, include_files):
    """
    Parse a DAG from a YAML workflow file.
    """
    with open(fname) as f:
        wf = yaml.load(f)

    dag = DAG()

    for job in wf["jobs"]:
        # parse job
        j = Job()

        # compute job
        if job["type"] == "job":
            j.xform = job["name"]
        # subworkflow job
        else:
            j.xform = job["file"]

        j.id = j.label = job["id"]
        dag.nodes[j.id] = j

        if job.get("nodeLabel"):
            j.label = job.get("nodeLabel")
        else:
            j.label = ""

        # parse uses (files)
        if include_files:
            for use in job["uses"]:
                if use["lfn"] in dag.nodes:
                    f = dag.nodes[use["lfn"]]
                else:
                    f = File()
                    f.id = f.label = use["lfn"]
                    dag.nodes[f.id] = f

                link_type = use["type"]

                if link_type == "input":
                    j.parents.append(f)
                    f.children.append(j)
                elif link_type == "output" or link_type == "checkpoint":
                    j.children.append(f)
                    f.parents.append(j)
                elif link_type == "inout":
                    print(
                        "WARNING: inout file {} of {} creates a cycle.".format(
                            f.id, j.id
                        )
                    )
                    f.children.append(j)
                    f.parents.append(j)
                    j.parents.append(f)
                    j.children.append(f)
                elif link_type == "none":
                    pass
                else:
                    raise Exception("Unrecognized link value: {}".format(link_type))

    for dep in wf["jobDependencies"]:
        for child in dep["children"]:
            dag.nodes[dep["id"]].children.append(dag.nodes[child])
            dag.nodes[child].parents.append(dag.nodes[dep["id"]])

    return dag


def remove_xforms(dag, xforms):
    """
    Remove transformations in the DAG by name
    """
    nodes = dag.nodes
    if len(xforms) == 0:
        return

    to_delete = []
    for _id in nodes.keys():
        node = nodes[_id]
        if isinstance(node, Job) and node.xform in xforms:
            print("Removing %s" % node.id)
            for p in node.parents:
                p.children.remove(node)
            for c in node.children:
                c.parents.remove(node)
            to_delete.append(_id)

    for _id in to_delete:
        del nodes[_id]


def transitivereduction(dag):
    # Perform a transitive reduction of the DAG to remove redundant edges.

    # First, perform a topological sort of the workflow.
    roots = [n for n in dag.nodes.values() if len(n.parents) == 0]

    L = []

    def visit(n):
        if n.mark == 1:
            raise Exception(
                "Workflow is not a DAG: Node %s is part of a "
                "cycle. Try without -f or with -s." % n
            )

        if n.mark == 0:
            n.mark = 1
            for m in n.children:
                visit(m)
            n.mark = 2
            L.insert(0, n)

    # Visit all the roots to create the topo sort
    for r in roots:
        visit(r)

    # Number all the levels of the workflow, which are used
    # to sort the children of each node in topological order.
    for n in L:
        n.level = 0
        for p in n.parents:
            n.level = max(n.level, p.level + 1)

    # The topological sort has to be reversed so that the deepest
    # nodes are visited first
    L.reverse()

    # This compares nodes by level for sorting. Note that sorting
    # children has to be done after the topo sort above because
    # the levels haven't been set until all the roots have been
    # visited.
    def lvlcmp(a, b):
        return a.level - b.level

    # This algorithm is due to Goralcikova and Koubek. It is fast and
    # simple, but it takes a lot of memory for large workflows because
    # it computes and stores the transitive closure of each node.
    for v in L:
        # This is to keep track of how many times v has been visited
        # from one of its parents. When this counter reaches the
        # number of parents the node has, then we can remove the closure
        v.mark = 0

        v.closure = {v}

        # We need to sort the children in topological order, otherwise the
        # reduction won't work properly. Sorting by level should produce
        # a valid topological ordering.
        v.children.sort(key=cmp_to_key(lvlcmp))

        # Compute the transitive closure and identify redundant edges
        reduced = []
        for w in v.children:

            w.mark += 1

            if isinstance(w, Job) and isinstance(v, Job) and w in v.closure:
                # If w is a Job, v is a Job, and w is in the closure, then it is not needed.
                # The above condition prevents us from removing edges from file -> job as those
                # edges should always be visible.
                sys.stderr.write("Removing {} -> {}\n".format(v.label, w.label))
            else:
                v.closure = v.closure.union(w.closure)
                reduced.append(w)

            # Once w has been visited by all its parents we can clear
            # its closure.
            if len(w.parents) == w.mark:
                w.closure = None

        # Another optimization. If v has no parents, then
        # we don't need to save its closure at all.
        if len(v.parents) == 0:
            v.closure = None

        # Now remove the edges
        v.children = reduced

    return dag


class emit_dot:
    """Write a DOT-formatted diagram.
    Options:
        label_type: What attribute to use for labels
        outfile: The file name to write the diagam out to.
        width: The width of the diagram
        height: The height of the diagram
    """

    def __init__(
        self, dag, label_type="label", outfile="/dev/stdout", width=None, height=None
    ):
        self.label_type = label_type

        self.next_color = 0  # Keep track of next color
        self.colors = {}  # Keep track of transformation names to assign colors

        self.out = open(outfile, "w")
        # Render the header
        self.out.write("digraph dag {\n")
        if width and height:
            self.out.write('    size="{:0.1f},{:0.1f}"\n'.format(width, height))
        self.out.write("    ratio=fill\n")
        self.out.write('    node [style=filled,color="#444444",fillcolor="#ffed6f"]\n')
        self.out.write("    edge [arrowhead=normal,arrowsize=1.0]\n\n")

        # Ensure that dot rendered in a deterministic manner
        nodes = sorted(dag.nodes.values(), key=lambda n: n.id)

        # Render nodes
        for n in nodes:
            n.renderNode(self)

        # Render edges
        for p in nodes:
            children = sorted(p.children, key=lambda n: n.id)
            for c in children:
                c.renderEdge(self, p)

        self.out.write("}\n")
        self.out.close()

    def getcolor(self, item):
        "Get the color for xform"
        if item not in self.colors:
            self.colors[item] = COLORS[self.next_color]
            # We use the modulus just in case we run out of colors
            self.next_color = (self.next_color + 1) % len(COLORS)
        return self.colors[item]

    def renderNode(self, id, label, fillcolor, color="#000000", shape="ellipse"):
        self.out.write(
            '    "%s" [shape=%s,color="%s",fillcolor="%s",label="%s"]\n'
            % (id, shape, color, fillcolor, label)
        )

    def renderEdge(self, parentid, childid, color="#000000"):
        self.out.write(
            '    "{}" -> "{}" [color="{}"]\n'.format(parentid, childid, color)
        )


def invoke_dot(dot_file, fmt, output):
    dot = shutil.which("dot")
    if dot:
        cmd = [dot]
        # output format
        cmd.append("-T{}".format(fmt))

        # output file
        cmd.extend(["-o", output])

        # path to dot file
        cmd.append(dot_file)

        subprocess.run(cmd)
    else:
        raise RuntimeError(
            "Unable to find 'dot'. Please install graphviz and ensure that 'dot' is added to your PATH"
        )


SUPPORTED_DRAW_FORMATS = {"jpg", "jpeg", "png", "pdf", "gif", "svg"}


def main():
    labeloptions = ["label", "xform", "id", "xform-id", "label-xform", "label-id"]
    labeloptionsstring = ", ".join("'%s'" % l for l in labeloptions)
    usage = "%prog [options] FILE"
    description = """Parses FILE and generates a DOT-formatted
graphical representation of the DAG. FILE can be a Condor
DAGMan file, Pegasus YAML file, or Pegasus DAX file."""
    parser = OptionParser(usage=usage, description=description)
    parser.add_option(
        "-s",
        "--nosimplify",
        action="store_false",
        dest="simplify",
        default=True,
        help="Do not simplify the graph by removing redundant edges. [default: False]",
    )
    parser.add_option(
        "-l",
        "--label",
        action="store",
        dest="label",
        default="label",
        help="What attribute to use for labels. One of %s. "
        "For 'label', the transformation is used for jobs that have no node-label. "
        "[default: label]" % labeloptionsstring,
    )
    parser.add_option(
        "-o",
        "--output",
        action="store",
        dest="outfile",
        metavar="FILE",
        default="/dev/stdout",
        help="""Write output to FILE [default: stdout]. If FILE is given with any
of the following extensions: 'png', 'jpg', 'jpeg', 'pdf', 'gif', and 'svg', pegasus-graphviz
will internally invoke 'dot -T<extension> -o FILE'. Note that graphviz must be
installed to output these file types. If any other extension is given, the raw
dot representation is output.""",
    )
    parser.add_option(
        "-r",
        "--remove",
        action="append",
        dest="remove",
        metavar="XFORM",
        default=[],
        help="Remove jobs from the workflow by transformation name. For subworkflows, use the workflow file name.",
    )
    parser.add_option(
        "-W",
        "--width",
        action="store",
        dest="width",
        type="float",
        default=None,
        help="Width of the digraph",
    )
    parser.add_option(
        "-H",
        "--height",
        action="store",
        dest="height",
        type="float",
        default=None,
        help="Height of the digraph",
    )
    parser.add_option(
        "-f",
        "--files",
        action="store_true",
        dest="files",
        default=False,
        help="Include files. This option is only valid for YAML and DAX files. [default: false]",
    )

    (options, args) = parser.parse_args()

    if options.width and options.height:
        pass
    elif options.width or options.height:
        parser.error("Either both --width and --height or neither")

    if options.label not in labeloptions:
        parser.error("--label must be one of %s" % labeloptionsstring)

    if len(args) < 1:
        parser.error("Please specify FILE")

    if len(args) > 1:
        parser.error("Invalid argument")

    dagfile = args[0]
    if dagfile.endswith(".dag"):
        dag = parse_dagfile(dagfile)
    elif dagfile.endswith(".dax") or dagfile.endswith(".xml"):
        dag = parse_daxfile(dagfile, options.files)
    elif dagfile.lower().endswith(".yml"):
        dag = parse_yamlfile(dagfile, options.files)
    else:
        raise RuntimeError(
            "Unrecognizable file format. Acceptable formats are '.dag', '.dax', '.xml', '.yml'"
        )

    remove_xforms(dag, options.remove)

    if options.simplify:
        dag = transitivereduction(dag)

    output_extension = Path(options.outfile).suffix.lower()[1:]
    if output_extension in SUPPORTED_DRAW_FORMATS:
        with NamedTemporaryFile("w") as f:
            emit_dot(dag, options.label, f.name, options.width, options.height)
            try:
                invoke_dot(f.name, output_extension, options.outfile)
            except RuntimeError as e:
                print("ERROR: {}".format(e))
                sys.exit(1)
    else:
        emit_dot(dag, options.label, options.outfile, options.width, options.height)


if __name__ == "__main__":
    main()



import json
import logging
import os
import re
import subprocess
from pathlib import Path

from flask import g, make_response, request, url_for

from Pegasus.db import connection
from Pegasus.db.ensembles import (
    EMError,
    Ensembles,
    EnsembleWorkflowStates,
    Triggers,
    TriggerType,
)
from Pegasus.service.ensembles import api, emapp
from Pegasus.service.lifecycle import authenticate

log = logging.getLogger(__name__)


def connect():
    log.debug("Connecting to database")
    g.master_db_url = g.user.get_master_db_url()
    g.session = connection.connect(
        g.master_db_url, connect_args={"check_same_thread": False}
    )


def disconnect():
    if "conn" in g:
        log.debug("Disconnecting from database")
        g.session.close()


@emapp.errorhandler(Exception)
def handle_error(e):
    return api.json_api_error(e)


emapp.before_request(authenticate)
emapp.before_request(connect)


@emapp.teardown_request
def teardown_request(exception):
    disconnect()


@emapp.route("/ensembles", methods=["GET"])
def route_list_ensembles():
    dao = Ensembles(g.session)
    ensembles = dao.list_ensembles(g.user.username)
    result = [e.get_object() for e in ensembles]
    return api.json_response(result)


@emapp.route("/ensembles", methods=["POST"])
def route_create_ensemble():
    name = request.form.get("name", None)
    if name is None:
        raise EMError("Specify ensemble name")

    max_running = request.form.get("max_running", 1)
    max_planning = request.form.get("max_planning", 1)

    dao = Ensembles(g.session)
    dao.create_ensemble(g.user.username, name, max_running, max_planning)
    g.session.commit()

    return api.json_created(url_for("route_get_ensemble", name=name, _external=True))


@emapp.route("/ensembles/<string:name>", methods=["GET"])
def route_get_ensemble(name):
    dao = Ensembles(g.session)
    e = dao.get_ensemble(g.user.username, name)
    result = e.get_object()
    return api.json_response(result)


@emapp.route("/ensembles/<string:name>", methods=["PUT", "POST"])
def route_update_ensemble(name):
    dao = Ensembles(g.session)
    e = dao.get_ensemble(g.user.username, name)

    max_running = request.form.get("max_running", None)
    if max_running is not None:
        e.set_max_running(max_running)

    max_planning = request.form.get("max_planning", None)
    if max_planning is not None:
        e.set_max_planning(max_planning)

    state = request.form.get("state", None)
    if state is not None:
        if state != e.state:
            # TODO Do the necessary state transition
            e.set_state(state)

    e.set_updated()

    g.session.commit()

    return api.json_response(e.get_object())


@emapp.route("/ensembles/<string:name>/workflows", methods=["GET"])
def route_list_ensemble_workflows(name):
    dao = Ensembles(g.session)
    e = dao.get_ensemble(g.user.username, name)
    result = [w.get_object() for w in dao.list_ensemble_workflows(e.id)]
    return api.json_response(result)


@emapp.route("/ensembles/<string:ensemble>/workflows", methods=["POST"])
def route_create_ensemble_workflow(ensemble):
    dao = Ensembles(g.session)
    e = dao.get_ensemble(g.user.username, ensemble)

    name = request.form.get("name", None)
    if name is None:
        raise EMError("Specify ensemble workflow 'name'")

    priority = request.form.get("priority", 0)

    basedir = request.form.get("basedir")
    if basedir is None:
        raise EMError("Specify 'basedir' where plan command should be executed")

    plan_command = request.form.get("plan_command")
    if plan_command is None:
        raise EMError("Specify 'plan_command' that should be executed to plan workflow")

    dao.create_ensemble_workflow(e.id, name, basedir, priority, plan_command)

    g.session.commit()

    return api.json_created(
        url_for("route_get_ensemble_workflow", ensemble=ensemble, workflow=name)
    )


@emapp.route(
    "/ensembles/<string:ensemble>/workflows/<string:workflow>", methods=["GET"]
)
def route_get_ensemble_workflow(ensemble, workflow):
    dao = Ensembles(g.session)
    e = dao.get_ensemble(g.user.username, ensemble)
    w = dao.get_ensemble_workflow(e.id, workflow)
    result = w.get_detail_object()
    return api.json_response(result)


@emapp.route(
    "/ensembles/<string:ensemble>/workflows/<string:workflow>", methods=["PUT", "POST"]
)
def route_update_ensemble_workflow(ensemble, workflow):
    dao = Ensembles(g.session)

    e = dao.get_ensemble(g.user.username, ensemble)
    w = dao.get_ensemble_workflow(e.id, workflow)

    priority = request.form.get("priority", None)
    if priority is not None:
        w.set_priority(priority)

    state = request.form.get("state", None)
    if state is not None:
        w.change_state(state)

    w.set_updated()

    g.session.commit()

    return api.json_response(w.get_detail_object())


@emapp.route(
    "/ensembles/<string:ensemble>/workflows/<string:workflow>/analyze", methods=["GET"]
)
def route_analyze_ensemble_workflow(ensemble, workflow):
    dao = Ensembles(g.session)
    e = dao.get_ensemble(g.user.username, ensemble)
    w = dao.get_ensemble_workflow(e.id, workflow)
    report = "".join(analyze(w))
    resp = make_response(report, 200)
    resp.headers["Content-Type"] = "text/plain"
    return resp


def analyze(workflow):
    w = workflow

    yield "Workflow state is %s\n" % w.state
    yield "Plan command is: %s\n" % w.plan_command

    logfile = w.get_logfile()
    if os.path.isfile(logfile):
        yield "Workflow log:\n"
        for l in open(w.get_logfile(), "rb"):
            yield "LOG: %s" % l.decode()
    else:
        yield "No workflow log available\n"

    if w.submitdir is None or not os.path.isdir(w.submitdir):
        yield "No submit directory available\n"
    else:
        yield "pegasus-analyzer output is:\n"
        p = subprocess.Popen(
            ["pegasus-analyzer", w.submitdir],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        out, err = p.communicate()
        out = out.decode()

        for l in out.split("\n"):
            yield "ANALYZER: %s\n" % l
        rc = p.wait()
        yield "ANALYZER: Exited with code %d\n" % rc

    if w.state == EnsembleWorkflowStates.PLAN_FAILED:
        yield "Planner failure detected\n"
    elif w.state == EnsembleWorkflowStates.RUN_FAILED:
        yield "pegasus-run failure detected\n"
    elif w.state == EnsembleWorkflowStates.FAILED:
        yield "Workflow failure detected\n"


# --- trigger related routes ---------------------------------------------------
@emapp.route("/ensembles/<string:ensemble>/triggers", methods=["GET"])
def route_list_triggers(ensemble):
    dao = Triggers(g.session)
    triggers = dao.list_triggers_by_ensemble(g.user.username, ensemble)

    return api.json_response([Triggers.get_object(t) for t in triggers])


@emapp.route("/ensembles/<string:ensemble>/triggers/<string:trigger>", methods=["GET"])
def route_get_trigger(ensemble, trigger):
    raise NotImplementedError("TODO")


# TODO: checks for correct data should be done here on the backend
"""
# error response format

{
  id: "id", <-- unique id to a request, it has been added as request.uid (use this when logging)
  code: "UNPROCESSABLE_ENTITY", <-- capitalized versions of errors that json schema would return 
  "message": "Err description",
  "errors": [
    {
      code: "MIN_LEN_ERR",
      message": "Err description",
      path: [ field_name ],
    },
    ..
  ],
  "warnings": [ .. ]
}
"""


@emapp.route("/ensembles/<string:ensemble>/triggers/cron", methods=["POST"])
def route_create_cron_trigger(ensemble):
    # verify that ensemble exists for user
    e_dao = Ensembles(g.session)

    # raises EMError code 404 if does not exist
    ensemble_id = e_dao.get_ensemble(g.user.username, ensemble).id

    # validate trigger
    trigger = request.form.get("trigger", type=str)
    if not trigger or len(trigger) == 0:
        raise EMError("trigger name must be a non-empty string")

    # validate workflow_script
    workflow_script = request.form.get("workflow_script", type=str)
    if not workflow_script or len(workflow_script) == 0:
        raise EMError("workflow_script name must be a non-empty string")

    if not Path(workflow_script).is_absolute():
        raise EMError("workflow_script must be given as an absolute path")

    # validate workflow_args
    can_decode = True
    try:
        workflow_args = json.loads(request.form.get("workflow_args"))
    except json.JSONDecodeError:
        can_decode = False

    if not can_decode or not isinstance(workflow_args, list):
        raise EMError("workflow_args must be given as a list serialized to json")

    # validate interval
    try:
        interval = to_seconds(request.form.get("interval", type=str))
    except ValueError:
        raise EMError(
            "interval must be given as `<int> <s|m|h|d>` and be greater than 0 seconds"
        )

    # validate timeout
    try:
        timeout = request.form.get("timeout", type=str, default=None)
        if timeout is not None:
            timeout = to_seconds(timeout)
    except ValueError:
        raise EMError(
            "timeout must be given as `<int> <s|m|h|d>` and be greater than 0 seconds"
        )

    kwargs = {
        "ensemble_id": ensemble_id,
        "trigger": trigger,
        "trigger_type": TriggerType.CRON.value,
        "workflow_script": workflow_script,
        "workflow_args": workflow_args,
        "interval": interval,
        "timeout": timeout,
    }

    # create trigger entry in db
    t_dao = Triggers(g.session)
    t_dao.insert_trigger(**kwargs)

    # return response success
    return api.json_created(
        url_for("route_get_trigger", ensemble=ensemble, trigger=trigger)
    )


@emapp.route("/ensembles/<string:ensemble>/triggers/file_pattern", methods=["POST"])
def route_create_file_pattern_trigger(ensemble):
    # verify that ensemble exists for user
    e_dao = Ensembles(g.session)

    # raises EMError code 404 if does not exist
    ensemble_id = e_dao.get_ensemble(g.user.username, ensemble).id

    # validate trigger
    trigger = request.form.get("trigger", type=str)
    if not trigger or len(trigger) == 0:
        raise EMError("trigger name must be a non-empty string")

    # validate workflow_script
    workflow_script = request.form.get("workflow_script", type=str)
    if not workflow_script or len(workflow_script) == 0:
        raise EMError("workflow_script name must be a non-empty string")

    if not Path(workflow_script).is_absolute():
        raise EMError("workflow_script must be given as an absolute path")

    # validate workflow_args
    can_decode = True
    try:
        workflow_args = json.loads(request.form.get("workflow_args"))
    except json.JSONDecodeError:
        can_decode = False

    if not can_decode or not isinstance(workflow_args, list):
        raise EMError("workflow_args must be given as a list serialized to json")

    # validate interval
    try:
        interval = to_seconds(request.form.get("interval", type=str))
    except ValueError:
        raise EMError(
            "interval must be given as `<int> <s|m|h|d>` and be greater than 0 seconds"
        )

    # validate timeout
    try:
        timeout = request.form.get("timeout", type=str, default=None)
        if timeout is not None:
            timeout = to_seconds(timeout)
    except ValueError:
        raise EMError(
            "timeout must be given as `<int> <s|m|h|d>` and be greater than 0 seconds"
        )

    # validate file_patterns
    can_decode = True
    try:
        file_patterns = json.loads(request.form.get("file_patterns"))
    except json.JSONDecodeError:
        can_decode = False

    if not can_decode or not isinstance(file_patterns, list):
        raise EMError("file_patterns must be given as a list serialized to json")

    if len(file_patterns) < 1:
        raise EMError("file_patterns must contain at least one file pattern")

    for fp in file_patterns:
        if not Path(fp).is_absolute():
            raise EMError(
                "each file pattern must be given as an absolute path (e.g. '/inputs/*.txt"
            )

    kwargs = {
        "ensemble_id": ensemble_id,
        "trigger": trigger,
        "trigger_type": TriggerType.FILE_PATTERN.value,
        "workflow_script": workflow_script,
        "workflow_args": workflow_args,
        "interval": interval,
        "timeout": timeout,
        "file_patterns": file_patterns,
    }

    # create trigger entry in db
    t_dao = Triggers(g.session)
    t_dao.insert_trigger(**kwargs)

    # return response success
    return api.json_created(
        url_for("route_get_trigger", ensemble=ensemble, trigger=trigger)
    )


@emapp.route(
    "/ensembles/<string:ensemble>/triggers/<string:trigger>", methods=["DELETE"]
)
def route_delete_trigger(ensemble, trigger):
    # verify that ensemble exists for user
    e_dao = Ensembles(g.session)

    # raises EMError code 404 if does not exist
    ensemble_id = e_dao.get_ensemble(g.user.username, ensemble).id

    # update trigger state to be STOPPED so that the TriggerManager can
    # handle it appropriately
    t_dao = Triggers(g.session)

    # make sure get_trigger raises 404 if nothing found
    trigger_id = t_dao.get_trigger(ensemble_id, trigger)._id
    t_dao.update_state(ensemble_id, trigger_id, "STOPPED")

    return api.json_response(
        {
            "message": "ensemble: {}, trigger: {} marked for deletion".format(
                ensemble, trigger
            )
        },
        status_code=202,
    )


def to_seconds(value: str) -> int:
    """Convert time unit given as '<int> <s|m|h|d>` to seconds.
    :param value: input str
    :type value: str
    :raises ValueError: value must be given as '<int> <s|m|h|d>
    :raises ValueError: value must be > 0s
    :return: value given in seconds
    :rtype: int
    """

    value = value.strip()
    pattern = re.compile(r"\d+ *[sSmMhHdD]")
    if not pattern.fullmatch(value):
        raise ValueError(
            "invalid interval: {}, interval must be given as '<int> <s|m|h|d>'".format(
                value
            )
        )

    num = int(value[0 : len(value) - 1])
    unit = value[-1].lower()

    as_seconds = {"s": 1, "m": 60, "h": 60 * 60, "d": 60 * 60 * 24}

    result = as_seconds[unit] * num

    if result <= 0:
        raise ValueError(
            "invalid interval: {}, interval must be greater than 0 seconds".format(
                result
            )
        )

    return result



#!/usr/bin/env python3

"""
Pegasus utility for checking that prereqs are installed
"""

import sys

if not sys.version_info >= (3, 5):
    sys.stderr.write("Pegasus requires Python 3.5 or above\n")
    sys.exit(1)

try:
    import yaml  # noqa
except Exception:
    sys.stderr.write("Pegasus requires the Python3 YAML module to be installed.\n")
    sys.stderr.write(
        "Please install it via your OS package manager, via pip, or activate a venv which includes the module.\n"
    )
    sys.exit(1)

try:
    import git  # noqa
except Exception:
    sys.stderr.write("Pegasus requires the Python3 GitPython module to be installed\n")
    sys.stderr.write(
        "Please install it via your OS package manager, via pip, or activate a venv which includes the module.\n"
    )
    sys.exit(1)



import logging
import os
import random

import click
import flask
from OpenSSL import crypto
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from werkzeug.serving import run_simple

from Pegasus.service import cache
from Pegasus.service._encoder import PegasusJsonEncoder
from Pegasus.service.base import BooleanConverter
from Pegasus.service.filters import register_jinja2_filters
from Pegasus.service.lifecycle import register_lifecycle_handlers

log = logging.getLogger(__name__)

# Services
services = ["dashboard", "monitoring"]


def generate_self_signed_certificate(certfile, pkeyfile):
    """
    SSL.
    :param certfile:
    :param pkeyfile:
    :return:
    If certfile and pkeyfile don't exist, create a self-signed certificate
    """

    if os.path.isfile(certfile) and os.path.isfile(pkeyfile):
        return

    logging.info("Generating self-signed certificate")

    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 2048)

    cert = crypto.X509()

    sub = cert.get_subject()
    sub.C = "US"
    sub.ST = "California"
    sub.L = "Marina Del Rey"
    sub.O = "University of Southern California"
    sub.OU = "Information Sciences Institute"
    sub.CN = "Pegasus Service"

    cert.set_version(1)
    cert.set_serial_number(random.randint(0, 2 ** 32))
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)  # 10 years
    cert.set_issuer(sub)
    cert.set_pubkey(pkey)
    cert.sign(pkey, "sha1")

    open(certfile, "wb").write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
    open(pkeyfile, "wb").write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))


def run(host="localhost", port=5000, debug=True, verbose=logging.INFO, **kwargs):
    app = create_app(env=os.getenv("FLASK_ENV", "development"))

    if debug:
        app.config.update(DEBUG=True)
        logging.getLogger().setLevel(logging.DEBUG)

    pegasus_service_url_prefix = os.environ.get(
        "PEGASUS_SERVICE_URL_PREFIX", app.config.get("PEGASUS_SERVICE_URL_PREFIX", None)
    )
    if pegasus_service_url_prefix:
        logging.info(
            "Using non-standard URL prefix: {}".format(pegasus_service_url_prefix)
        )
        app.config["APPLICATION_ROOT"] = pegasus_service_url_prefix

    pegasusdir = os.path.expanduser("~/.pegasus")
    if not os.path.isdir(pegasusdir):
        os.makedirs(pegasusdir, mode=0o744)

    cert = app.config.get("CERTIFICATE", None)
    pkey = app.config.get("PRIVATE_KEY", None)
    if cert is None or pkey is None:
        log.warning("SSL is not configured: Using self-signed certificate")
        cert = os.path.expanduser("~/.pegasus/selfcert.pem")
        pkey = os.path.expanduser("~/.pegasus/selfkey.pem")
        generate_self_signed_certificate(cert, pkey)
    ssl_context = (cert, pkey)

    if os.getuid() != 0:
        log.warning("Service not running as root: Will not be able to switch users")

    options = {}
    if app.config.get("MAX_PROCESSES", None):
        options = {
            "threaded": False,
            "processes": app.config["MAX_PROCESSES"],
            "use_reloader": False,
        }
    else:
        options = {
            "threaded": True,
            "use_reloader": True,
        }

    if pegasus_service_url_prefix:
        application = DispatcherMiddleware(
            flask.Flask("dummy_app"), {app.config["APPLICATION_ROOT"]: app}
        )
        run_simple(
            host, port, application, ssl_context=ssl_context, **options,
        )
    else:
        app.run(
            host=host, port=port, ssl_context=ssl_context, **options,
        )

    log.info("Exiting")


def _load_user_config(app):
    # Load user configuration
    conf = os.path.expanduser("~/.pegasus/service.py")
    if os.path.isfile(conf):
        app.config.from_pyfile(conf)


def create_app(config=None, env="development"):
    """Configure app."""
    # Environment
    os.environ["FLASK_ENV"] = env

    app = flask.Flask(__name__)

    # Flask Configuration
    app.config.from_object("Pegasus.service.defaults")
    # app.config.from_object("Pegasus.service.config.%sConfig" % env.capitalize())
    _load_user_config(app)
    app.config.update(config or {})

    if "PEGASUS_ENV" in os.environ:
        app.config.from_envvar("PEGASUS_ENV")

    # Initialize Extensions
    cache.init_app(app)
    # db.init_app(app)
    # socketio.init_app(app, json=flask.json)

    configure_app(app)

    # Service Configuration
    for service in services:
        config_method = "configure_%s" % service
        if config_method in globals():
            globals()["configure_%s" % service](app)

    return app


def configure_app(app):
    #
    # Flask URL variables support int, float, and path converters.
    # Adding support for a boolean converter.
    #
    app.url_map.converters["boolean"] = BooleanConverter

    #
    # Relax trailing slash requirement
    #
    app.url_map.strict_slashes = False

    # Attach global JSONEncoder
    app.json_encoder = PegasusJsonEncoder

    # Register lifecycle methods
    register_lifecycle_handlers(app)

    # Register Jinja2 Filters
    register_jinja2_filters(app)

    # Error handlers
    ## register_error_handlers(app)
    ...


def configure_dashboard(app):
    from Pegasus.service.dashboard import blueprint

    app.register_blueprint(blueprint)


def configure_monitoring(app):
    from Pegasus.service.monitoring import monitoring

    app.register_blueprint(monitoring, url_prefix="/api/v1/user/<string:username>")


@click.command(name="pegasus-service")
@click.option(
    "--host",
    default="localhost",
    metavar="<hostname>",
    show_default=True,
    help="Hostname",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=5000,
    metavar="<port-number>",
    show_default=True,
    help="Port no. on which to listen for requests",
)
@click.option(
    "-d/-nd",
    "--debug/--no-debug",
    default=True,
    metavar="<debug-mode>",
    help="Start server in development mode",
)
@click.option(
    "-v",
    "--verbose",
    default=logging.DEBUG,
    count=True,
    metavar="<verbosity>",
    help="Logging verbosity",
)
def main(host: str, port: int, debug: bool, verbose: int):
    """Run the Pegasus Service server."""
    run(host=host, port=port, debug=debug, verbose=verbose)


if __name__ == "__main__":
    main()
